/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.io.*;

public abstract class Encryption {


    //
    // flags
    //

    static Boolean supported;

    public static boolean isSupported () {
        if (supported == null) {
            String ver = System.getProperty ("java.version");
            if (ver.startsWith ("1.0") 
                || ver.startsWith ("1.1"))
                supported = Boolean.FALSE;
            else {
                try {
                    Class.forName ("javax.net.ssl.SSLSocket");
                    supported = Boolean.TRUE;
                } catch (ClassNotFoundException e) {
                    supported = Boolean.FALSE;
                }
            }
        }
        return supported.booleanValue ();
    }

    public static boolean isRequired () {
        return isSupported ()
            && 
            Boolean.valueOf 
            (Global.getConfig ().getProperty ("EncryptionRequired", "true"))
            .booleanValue ();
    }

    public static void setRequired (boolean yes) {
        Global.getConfig ().put ("EncryptionRequired", yes ? "true" : "false");
        Global.getConfig ().save ();
    }


    public static boolean isRequiredLocally () {
        return isSupported ()
            && 
            Boolean.valueOf 
            (Global.getConfig ().getProperty ("EncryptionRequiredLocally", 
                                              "true"))
            .booleanValue ();
    }

    public static void setRequiredLocally (boolean yes) {
        Global.getConfig ().put ("EncryptionRequiredLocally", yes ? "true" : "false");
        Global.getConfig ().save ();
    }


    //
    // singleton instance
    //

    static Encryption instance;

    public static Encryption getInstance () {
        if (instance == null) {
            try {
                if (isSupported ())
                    instance = 
                        (Encryption)
                        Class.forName ("SSLEncryption").newInstance ();
            } catch (Exception e) {
                Global.report (e);
            }
            if (instance == null)
                instance = new NoEncryption ();

                Cert[] certs = instance.getCertificates ();
                for (int i = 0; i < certs.length; ++i)
                    Global.debug (certs[i].getName () + " (" 
                                  + certs[i].getFingerprint ()
                                  + ")");
                
        }
        return instance;            
    }

    //
    // Keystore filename
    //
    
    static File keystoreFile =
        new File (Global.getConfig ()
                  .getProperty ("Keystore", 
                                new File (Global.getHomeDirectory (),
                                          ".rclip-keys").toString ()));

    public static File getKeystoreFile () {
        return keystoreFile;
    }

    public static void setKeystoreFile (File f) {
        keystoreFile = f;
    }


    //
    // Managing keystore
    //
        
    public abstract void regenerateKeys () throws Exception;
    
    public abstract void reloadKeys () throws Exception;

    /**
     * List the certificates in the keystore.
     * @return array of Encryption.Certificate objects.
     */
    public abstract Cert[] getCertificates ();

    public static interface Cert {
        String getName ();
        String getFingerprint ();
        void delete ();

    }
}
