/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.io.*;
import java.util.*;
import rcm.progress.*;

public class FileList implements Serializable {
    static final long serialVersionUID = -653239533042926902L;
    protected static final int BUFFER_SIZE = 100 * 1024;

    Vector names = new Vector (); // Vector of Strings (base names) 
    Vector files = new Vector (); // Vector of Files (absolute paths)

    public FileList () {
    }

    public FileList (String[] f) throws IOException {
        for (int i=0; i < f.length; ++i)
            add (f[i]);
    }

    public void add (String filename) throws IOException {
        add (new File (filename));
    }

    public void add (File file) throws IOException {
        if (!file.exists ())
            throw new FileNotFoundException (file.toString ());
        if (!file.isAbsolute ())
            file = new File (file.getAbsolutePath ());
        files.addElement (file);
        names.addElement (file.getName ());
    }

    public String[] getNames () {
        int n = length ();
        String[] namesArray = new String[length ()];
        for (int i = 0; i < n; ++i)
            namesArray[i] = names.elementAt (i).toString ();
        return namesArray;
    }

    public String[] getPathnames () {
        int n = length ();
        String[] names = new String[n];
        for (int i = 0; i < n; ++i)
            names[i] = ((File)files.elementAt (i)).getAbsolutePath ();
        return names;
    }

    public boolean isEmpty () {
        return files.isEmpty ();
    }

    public int length () {
        return files.size ();
    }

    public void copy (String directory) throws IOException {
        copy (directory, null);
    }

    public void copy (String directory, 
                      ProgressListener listener) throws IOException {
        byte[] buffer = new byte[BUFFER_SIZE];
        
        BasicProgressGenerator gen = new BasicProgressGenerator (this);
        if (listener != null)
            gen.addProgressListener (listener);
            
        // Make sure dir has a terminating separator
        if (directory == null)
            directory = ".";
        else if (directory.length () > 0
            && !(directory.endsWith ("/") ||
                 directory.endsWith (File.separator)))
            directory += File.separator;

        File targetDir = new File (directory);

        for (Enumeration e = files.elements (); e.hasMoreElements ();)
            copyFile ((File)e.nextElement (), targetDir, buffer, gen);
    }

    private void copyFile (File contents, File targetDir, byte[] buffer, 
                           BasicProgressGenerator gen) throws IOException {
        File target = new File (targetDir, contents.getName ());
        
        if (contents.isDirectory ()) {
            // Create a corresponding target directory
            gen.fireStartProgress (target.toString (), 1);
            target.mkdir ();
            gen.fireDoneProgress ();

            // Add each file in directory recursively
            String[] subfiles = contents.list ();
            for (int i = 0; i < subfiles.length; ++i)
                copyFile (new File (contents, subfiles[i]), 
                          target, buffer, gen);
        }
        else {
            // Copy the file
            long reads = (contents.length () + buffer.length - 1)
                / buffer.length;
            gen.fireStartProgress (target.toString (), (int)reads);

            InputStream in = new FileInputStream (contents);
            OutputStream out = new FileOutputStream (target);
            int n;
            while ((n = in.read (buffer)) != -1) {
                out.write (buffer, 0, n);
                gen.fireAddProgress (1);
            }
            in.close ();
            out.close ();
            gen.fireDoneProgress ();
        }
    }

    public void delete () {
        for (Enumeration e = files.elements (); e.hasMoreElements (); )
            deleteFile ((File)e.nextElement ());
    }

    public void deleteFile (File f) {
        if (f.isDirectory ()) {
            // Delete each file in directory recursively
            String[] subfiles = f.list ();
            for (int i = 0; i < subfiles.length; ++i)
                deleteFile (new File (f, subfiles[i]));
        }
        f.delete ();
    }

}
