/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.awt.*;
import java.awt.List;
import java.awt.event.*;
import java.net.*;
import java.io.*;
import java.util.*;
import java.awt.datatransfer.*;
import rcm.awt.*;
import rcm.progress.*;
import rcm.io.TempFile;
import rcm.util.Str;
import rcm.util.Win;
import rcm.win32.*;

public class GUI extends Frame implements UI {    
    SynchronizedClipboard clipboard; // my clipboard
    Locator locator;                 // my locator
    MetaProtocol protocol;

    Button addButton;
    Button removeButton;
    Button reconnectButton;
    Button optionsButton;
    Button exitButton;

    List groupList;     // group members

    Label fingerprint;  // this host's fingerprint (if any)

    Label status;           // status of last operation

    Button copyFilesButton;// copies multiple files to synchronized clipboard
    
    Button pasteFilesButton;  // pastes files from synchronized clipboard
    
    TrayIcon trayIcon;    // system tray icon on Win32 platforms, otherwise null

    public GUI (SynchronizedClipboard c, Locator locator, MetaProtocol protocol) throws Exception {
        super ("Remote Clip");

        if (Global.getDebug ().isEnabled ()) {
            setTitle ("Remote Clip ("
                      + locator.getAddress ().getHostAddress ()
                      + ":"
                      + locator.getPort ()
                      + ")");
        }

        this.clipboard = c;
        this.locator = locator;
        this.protocol = protocol;

        setLayout (new BorderLayout ());
        add (makeTopPanel (), BorderLayout.NORTH);
        add (makeGroupList (), BorderLayout.CENTER);
        add (makeBottomPanel (), BorderLayout.SOUTH);

        URL iconURL = getClass ().getResource ("icon.gif");
        Image icon = getToolkit ().getImage (iconURL);
        setIconImage (icon);

        try {
            trayIcon = new rcm.win32.TrayIcon ("Remote Clip");

            trayIcon.addMouseListener (new MouseAdapter () {
                    public void mouseClicked (MouseEvent evt) {
                        GUI.this.restore ();
                    }
                });

            trayIcon.setVisible (true);
        } catch (UnsupportedPlatformException e) {
            trayIcon = null;
        }

        addWindowListener (new WindowAdapter () {
            public void windowClosing (WindowEvent event) {
                if (trayIcon != null)
                    // closing the window doesn't exit Remote Clip,
                    // since the tray icon still exists
                    setVisible (false);
                else
                    exit (0);
            }
        });

        // Initialize the group list
        groupChanged ();

        pack ();
        show ();
        setLocation (100, 100);

        Global.debug ("Showed GUI");
    }

    void restore () {
        setVisible (true);
        setState (Frame.NORMAL);
        toFront ();
    }

    // code 0 is normal exit, code -1 is error exit, code 42 is restart
    void exit (int code) {
        try {
            if (trayIcon != null)
                trayIcon.dispose ();
            clipboard.close ();
            TempFile.deleteAllTempFiles ();
        } catch (Throwable t) {
            // Catch all exceptions to make sure we exit in any case 
            Global.report (t);
        }
        System.exit (code);                
    }

    Component makeTopPanel () {
        Panel panel = new Panel ();
        panel.setLayout (new FlowLayout (FlowLayout.LEFT, 2, 0));
        panel.add (new Label ("Clipboard Group:"));
        panel.add (addButton = new Button ("Add..."));
        addButton.addActionListener (new ActionListener () {
            public void actionPerformed (ActionEvent event) {
                new AddDialog ().show ();
            }
        });

        panel.add (removeButton = new Button ("Remove"));
        removeButton.setEnabled (false);
        removeButton.addActionListener (new ActionListener () {
            public void actionPerformed (ActionEvent event) {
                Clipboard[] clipboards = clipboard.getClipboards ();
                int[] indexes = groupList.getSelectedIndexes ();
                for (int i = 0; i < indexes.length; ++i) {
                    int k = indexes[i] - 1;
                    if (k >= 0) 
                        clipboard.remove (clipboards[k]);
                }
            }
        });

        panel.add (reconnectButton = new Button ("Reconnect"));
        reconnectButton.setEnabled (false);
        reconnectButton.addActionListener (new ActionListener () {
            public void actionPerformed (ActionEvent event) {
                Locator[] locators = clipboard.getLocators ();
                int[] indexes = groupList.getSelectedIndexes ();
                if (indexes != null)
                    for (int i = 0; i < indexes.length; ++i)
                        if (indexes[i] > 0) {
                            Locator l = locators[indexes[i]-1];
                            if (l != null)
                                clipboard.add (l);
                        }
                updateReconnectButton ();
            }
        });

        panel.add (optionsButton = new Button ("Options..."));
        optionsButton.addActionListener (new ActionListener () {
            public void actionPerformed (ActionEvent event) {
                new OptionsDialog ().show ();
            }
        });

        panel.add (exitButton = new Button ("Exit"));
        exitButton.addActionListener (new ActionListener () {
            public void actionPerformed (ActionEvent event) {
                exit (0);
            }
        });

        return panel;
    }

    Component makeGroupList () {
        groupList = new List (5);
        groupList.setMultipleMode (true);
        groupList.addItemListener (new ItemListener () {
            public void itemStateChanged (ItemEvent event) {
                updateRemoveButton ();
                updateReconnectButton ();
            }
        });
        return groupList;
    }

    Component makeBottomPanel () {
        Panel panel = new Panel ();
        panel.setLayout (new BorderLayout ());
        panel.add (makeFingerprint (), BorderLayout.NORTH);
        panel.add (makeTextHelp (), BorderLayout.WEST);
        panel.add (makeFilesHelp (), BorderLayout.EAST);
        panel.add (makeStatusBar (), BorderLayout.SOUTH);
        return panel;
    }

    Component makeFingerprint () {
        fingerprint = new Label ("");
        return BorderPanel.wrap (fingerprint, 2);
    }
        
    Component makeTextHelp () {
        Panel panel = new Panel ();
        panel.setLayout (new BorderLayout ());
        MultiLineLabel label = new MultiLineLabel
            ("To copy TEXT between computers,\n"
             + "use the Copy and Paste commands\n"
             + "in your Windows and X applications.");
        label.setFont (new Label().getFont ());
        panel.add (BorderPanel.wrap (label, 5), 
                   BorderLayout.NORTH);
        return panel;
    }

    Component makeFilesHelp () {
        Panel panel = new Panel ();
        panel.setLayout (new BorderLayout ());
        MultiLineLabel label = new MultiLineLabel
            ("To copy FILES between computers,\n"
             + "use either the Windows Explorer,\n"
             + "the rcopy and rpaste commands,\n"
             + "or the buttons below.");
        panel.add (BorderPanel.wrap (label, 5),
                   BorderLayout.NORTH);
        panel.add (makeFileButtons ());
        return panel;
    }


    Component makeFileButtons () {
        Panel panel = new Panel ();
        panel.setLayout (new FlowLayout (FlowLayout.CENTER, 5, 2));
        
        String homedir;
        if ((homedir = System.getProperty ("user.home")) != null
            || (homedir = System.getProperty ("user.dir")) != null)
            PopupDialog.currentDirectory = homedir;
            
        panel.add (copyFilesButton = new Button ("Copy Files"));
        copyFilesButton.addActionListener (new ActionListener () {
            public void actionPerformed (ActionEvent event) {
                new CopyFilesDialog ().show ();
            }
        });

        panel.add (pasteFilesButton = new Button ("Paste Files"));
        pasteFilesButton.setEnabled (false);
        pasteFilesButton.addActionListener (new ActionListener () {
            public void actionPerformed (ActionEvent event) {
                Transferable pasted = clipboard.getContents (null);
                try {
                    try {
                        FileList fl = (FileList) pasted.getTransferData 
                            (Flavors.filesFlavor);
                        new PasteFilesDialog (fl).show ();
                    } catch (UnsupportedFlavorException e1) {
                        reportStatus 
                            ("Paste failed: no files on clipboard.");
                    }
                } catch (IOException e) {
                    Global.report (e);
                    PopupDialog.warn (GUI.this, "Error", 
                                      e.getMessage ());
                    
                }
            }
        });

        return panel;
    }

    Component makeStatusBar () {
        // Status bar
        status = new Label 
            ("Ready.                                                      ");
        return status;
    }

    synchronized void updateRemoveButton () {
        int[] indexes = groupList.getSelectedIndexes ();
        // Enable it if at least one host is selected and
        // it's not localhost (index 0)
        removeButton.setEnabled (indexes != null
                                 && indexes.length > 0
                                 && !(indexes.length == 1
                                      && indexes[0] == 0));
    }
    
    synchronized void updateReconnectButton () {
        Clipboard[] clipboards = clipboard.getClipboards ();
        int[] indexes = groupList.getSelectedIndexes ();
        // Enable it if at least one non-localhost host is selected
        // which is disconnected or failed
        if (indexes != null) {
            for (int i = 0; i < indexes.length; ++i) {
                if (indexes[i] == 0)
                    // localhost, skip it
                    continue;
                if (clipboards[indexes[i] - 1] 
                    instanceof SynchronizedClipboard.DisconnectedClipboard) {
                    reconnectButton.setEnabled (true);
                    return;
                }
            }
        }
        
        reconnectButton.setEnabled (false);
    }
    
    //
    // status and error messages
    //
    public void reportStatus (String message) {
        status.setText (message);

        // Garbage-collect any temp files that are no longer referenced
        Runtime rt = Runtime.getRuntime ();
        rt.gc ();
        rt.runFinalization ();
    }

    public void setHourglass (boolean enabled) {
        Cursor cursor =
            Cursor.getPredefinedCursor (enabled 
                                        ? Cursor.WAIT_CURSOR
                                        : Cursor.DEFAULT_CURSOR);
        setCursor (cursor);
    }

    public void popupError (String message) {
        PopupDialog.warn (this, "Error", message);
    }

    public void setFingerprint (String message) {
        fingerprint.setText (message);
    }

    //
    // group change event handler
    //
    public synchronized void groupChanged () {
        groupList.removeAll ();
        groupList.add (clipboard.getName ());
        for (Enumeration e = clipboard.clipboards (); 
             e.hasMoreElements (); ) {
            Clipboard c = (Clipboard) e.nextElement ();
            groupList.add (c.getName ());
        }

        updateRemoveButton ();
        updateReconnectButton ();
    }

    //
    // current MIME type on clipboard
    //
    public void setMimeType (String type) {
        pasteFilesButton.setEnabled (Flavors.FILES_TYPE.equals (type));
    }

    //
    // asking user to authorize incoming connections
    //

    public boolean authorize (String message) {
        AuthorizationDialog dlg = 
            new AuthorizationDialog (message);
        dlg.show ();
        return (dlg.getAnswer() == RunnableDialog.YES);
    }

    class AuthorizationDialog extends RunnableDialog {
        public AuthorizationDialog (String message) {
            super (GUI.this, "Allow Clipboard Join?",
                   true,
                   message, null,
                   "Yes", "No", null);
            GUI.this.toFront ();
        }
        public void run () {
            // display dialog for 30 seconds, then assume NO
            try {
                Thread.sleep (30 * 1000); 
            } catch (InterruptedException e) {
            }
            setAnswer (NO);
        }
    }


    //
    // Dialog for adding host to clipboard group
    //
    class AddDialog extends OkCancelDialog {
        CheckboxGroup connectionType;
        Checkbox network;
        Checkbox serial;

        Label hostLabel;
        TextField host;
        Label portLabel;
        TextField port;
        Checkbox encrypted;
        
        Label serialPortLabel;
        TextField serialPort;

        public AddDialog () {
            super (GUI.this, "Add Host to Clipboard Group");

            Panel me = getControlsPane ();
            me.setLayout (new GridBagLayout ());

            connectionType = new CheckboxGroup ();
            network = new Checkbox (" Network", connectionType, true);
            serial = new Checkbox (" Serial", connectionType, false);

            me.add (hostLabel = new Label ("Hostname:"),
                    Constrain.labelLike (1, 1));
            me.add (host = new TextField (30), 
                    Constrain.fieldLike (2, 1, 2));

            me.add (portLabel = new Label ("Port:"),
                    Constrain.labelLike (1, 2));
            me.add (port = new TextField (String.valueOf (Locator.DEFAULT_PORT), 5),
                    Constrain.labelLike (2, 2));
            me.add (encrypted = new Checkbox ("Encrypted", true),
                    Constrain.labelLike (3, 2));

            if (!Encryption.isSupported ()) {
                encrypted.setState (false);
                encrypted.setEnabled (false);
            }

            if (Serial.isSupported ()) {
                me.add (network, Constrain.labelLike (0, 0)); 

                me.add (serial, Constrain.labelLike (0, 10)); 
                me.add (serialPortLabel = new Label ("Serial Port:"),
                        Constrain.labelLike (1, 11));
                me.add (serialPort = new TextField (20),
                        Constrain.labelLike (2, 11, 2));
                
                updateConnectionControls ();
            }
            
            ItemListener itemListener = new ItemListener () {
                    public void itemStateChanged (ItemEvent e) {
                        updateConnectionControls ();
                    }
                };
            network.addItemListener (itemListener);
            serial.addItemListener (itemListener);

            KeyListener keyListener = new KeyAdapter () {
                public void keyPressed (KeyEvent event) {
                    switch (event.getKeyCode ()) {
                        case KeyEvent.VK_ENTER:
                        ok ();
                        break;
                        case KeyEvent.VK_ESCAPE:
                        cancel ();
                        break;
                    }
                }
            };
            AddDialog.this.addKeyListener (keyListener);
            host.addKeyListener (keyListener);
            port.addKeyListener (keyListener);
            
            AddDialog.this.pack ();
        }

        void updateConnectionControls () {
            boolean f;

            f = network.getState ();
            hostLabel.setEnabled (f);
            host.setEnabled (f);
            portLabel.setEnabled (f);
            port.setEnabled (f);
            if (Encryption.isSupported ())
                encrypted.setEnabled (f);


            f = serial.getState ();
            serialPortLabel.setEnabled (f);
            serialPort.setEnabled (f);
        }                

        public void ok () {
            super.ok ();
            

            if (network.getState ()) {
                // add a network host
                String h = host.getText ().trim ();
                if (h.length () == 0)
                    return;
                
                String p = port.getText ().trim ();
                if (p.length () > 0)
                    p = ":" + p;
                
                String pr = 
                    Locator.PROTOCOL_PREFIX[(encrypted.getState ()) 
                                            ? Locator.SECURE_RMI
                                            : Locator.RMI];

                clipboard.add (new Locator (pr + h + p));
            } else {
                // add a serial port
                String p = serialPort.getText ().trim ();
                if (p.length () == 0)
                    return;

                String pr = Locator.PROTOCOL_PREFIX[Locator.SERIAL];

                clipboard.add (new Locator (pr + p));
            }

        }
    }

    // Copy Files dialog box
    
    class CopyFilesDialog extends OkCancelDialog {
        TextField editFilename; // current filename to edit
        List files;             // files on clipboard
        Button browseButton;    // pop up a file dialog
        Button addButton;       // adds a blank file to list
        Button removeButton;    // removes selected files from list

        public CopyFilesDialog () {
            super (Win.findFrame (GUI.this), 
                   "Choose Files or Directories To Copy");            

            Panel mainPanel = getControlsPane ();

            mainPanel.setLayout (new GridBagLayout ());

            Constrain.add (mainPanel, new Label ("Add:"), 
                           Constrain.labelLike (0, 0));

            Constrain.add (mainPanel, editFilename = new TextField (40), 
                           Constrain.fieldLike (1, 0));
            editFilename.addActionListener (new ActionListener () {
                public void actionPerformed (ActionEvent event) {
                    String fn = editFilename.getText ();
                    if (fn.length() == 0)
                        return;

                    if (!new File (fn).exists ()) {
                        PopupDialog.warn (GUI.this,
                                          "File Not Found",
                                          "File " + fn + " not found.");
                        editFilename.selectAll ();
                        return;
                    }
                     
                    files.add (fn);
                    editFilename.setText ("");
                }
            });

            Constrain.add (mainPanel, browseButton = new Button ("Browse..."),
                           Constrain.fieldLike (2, 0));
            browseButton.addActionListener (new ActionListener () {
                public void actionPerformed (ActionEvent event) {
                    String fn = PopupDialog.askFilename 
                        (GUI.this, 
                         "Choose File To Copy", "", true);
                    if (fn != null) {
                        files.add (fn);
                        editFilename.setText ("");
                    }
                }
            });
            
            Constrain.add (mainPanel, 
                           new Label ("Selected Files/Directories:"), 
                           Constrain.fieldLike (0, 1, 3));
            
            Constrain.add (mainPanel, files = new List (5), 
                           Constrain.areaLike (0, 2, 2));
            files.setMultipleMode (true);
            files.addItemListener (new ItemListener () {
                public void itemStateChanged (ItemEvent event) {
                    int[] indexes = files.getSelectedIndexes ();
                    removeButton.setEnabled (indexes != null
                                             && indexes.length > 0);
                }
            });


            Constrain.add (mainPanel, removeButton = new Button ("Remove"), 
                           Constrain.labelLike (2, 2));
            removeButton.setEnabled (false);
            removeButton.addActionListener (new ActionListener () {
                public void actionPerformed (ActionEvent event) {
                    int[] indexes = files.getSelectedIndexes ();
                    if (indexes != null)
                        for (int i = indexes.length-1; i >= 0; --i)
                            files.remove (indexes[i]);
                    editFilename.setText ("");
                    removeButton.setEnabled (false);
                }
            });
            
            CopyFilesDialog.this.pack ();
            PopupDialog.centerWindow (this, GUI.this);
        }
        
        public void ok () {
            super.ok ();            
            CopyFilesDialog.this.setVisible (false);
            try {
                Transferable contents = new ConstantTransferable 
                    (new FileList (files.getItems ()));
                clipboard.setContents (contents, null);
            } catch (IOException e) {
                popupError (e.toString ());
            }
        }
        
        public void cancel () {
            super.cancel ();
            reportStatus ("Copy cancelled.");
        }
        
    };

    // Paste Files dialog box
    
    class PasteFilesDialog extends OkCancelDialog implements Runnable {
        FileList fileList;     // pasted FileList

        List files;             // files on clipboard
        TextField saveTo;       // directory in which to store pasted files
        Button browseButton;    // pops up a directory chooser

        public PasteFilesDialog (FileList fileList) {
            this (fileList.getNames ());
            this.fileList = fileList;
            Global.debug ("file list is a " + fileList.getClass ());
        }

        private PasteFilesDialog (String[] filenames) {
            super (Win.findFrame (GUI.this), 
                   "Paste Files");            
            
            Panel mainPanel = getControlsPane ();

            mainPanel.setLayout (new GridBagLayout ());
            
            Constrain.add (mainPanel, new Label ("Pasted Files:"), 
                           Constrain.fieldLike (0, 0, 3));
            
            Constrain.add (mainPanel, files = new List (3, true), 
                           Constrain.areaLike (0, 1, 3));

            // Load listbox with filenames
            files.removeAll ();
            for (int i=0; i < filenames.length; ++i)
                files.add (filenames[i]);
                
            Constrain.add (mainPanel, new Label ("Paste Files To:"), 
                           Constrain.labelLike (0, 2));

            Constrain.add (mainPanel, 
                           saveTo = new TextField 
                           (PopupDialog.currentDirectory, 40), 
                           Constrain.fieldLike (1, 2));
            saveTo.addActionListener (new ActionListener () {
                public void actionPerformed (ActionEvent event) {
                    ok ();
                }
            });

            Constrain.add (mainPanel, browseButton = new Button ("Browse..."), 
                           Constrain.labelLike (2, 2));
            browseButton.addActionListener (new ActionListener () {
                public void actionPerformed (ActionEvent event) {
                    File defaultDir = new File (saveTo.getText ());
                    File defaultFn = new File (defaultDir, 
                        (files.getItemCount () > 0 ) ? files.getItem (0) : "");
                    String dir = 
                        PopupDialog.askDirectory 
                        (GUI.this, "Paste Files To Directory", 
                         defaultFn.toString (), false);
                    if (dir != null)
                        saveTo.setText (dir);
                }
            });
            
            PasteFilesDialog.this.pack ();
            PopupDialog.centerWindow (this, GUI.this);
        }

        AWTProgressBar progressBar;
        Dialog progressDialog;

        public void ok () {
            super.ok ();
            progressBar = new AWTProgressBar ();
            progressBar.getLabel ().setText (Str.repeat(" ", 60));
            progressDialog = progressBar.makeDialog (GUI.this, 
                                                     "Pasting Files", true);

            Thread t = new Thread (this);
            t.setPriority (Thread.MIN_PRIORITY);
            t.start ();

            progressDialog.show ();
        }

        public void cancel () {
            super.cancel ();
            reportStatus ("Paste cancelled.");
        }

        public void run () {
            String dir = saveTo.getText ();

            try {
                // Copy the files
                try {
                    fileList.copy (dir, progressBar);
                } finally { 
                    progressDialog.dispose ();
                }

                if (files.getItemCount() == 1)
                    reportStatus ("Pasted " + files.getItem(0) + 
                                  " to " + dir + ".");
                else
                    reportStatus ("Pasted " + files.getItemCount() + 
                                  " files to " + dir + ".");
            } catch (IOException e) {
                Global.report (e);
                PopupDialog.warn (GUI.this, 
                                  "Error", e.toString ());
                reportStatus ("Paste failed.");
            } catch (CancelException e) {
                reportStatus ("Paste cancelled.");
            }
        }
        
    };

    //
    // Options dialog
    //
    class OptionsDialog extends OkCancelDialog {
        TextField port;
        String originalPort;

        TextField keystoreFile;
        String originalKeystoreFile;
        Button browseButton;

        Checkbox encryptionRequired;
        Checkbox encryptionRequiredLocally;
        Button clearKeys;
        Button editKeys;

        public OptionsDialog () {
            super (GUI.this, "Remote Clip Options");

            Panel me = getControlsPane ();
            me.setLayout (new GridBagLayout ());

            
            originalPort = String.valueOf (Global.getLocalPort ());
            me.add (new Label ("Port:"),
                    Constrain.labelLike (0, 0));
            me.add (port = new TextField (originalPort),
                    Constrain.labelLike (1, 0));

            originalKeystoreFile = Encryption.getKeystoreFile ().toString ();
            me.add (new Label ("Keystore:"),
                    Constrain.labelLike (0, 1));
            me.add (keystoreFile = new TextField (originalKeystoreFile),
                    Constrain.fieldLike (1, 1, 2));
            me.add (browseButton = new Button ("..."),
                    Constrain.labelLike (2, 1));
            browseButton.addActionListener (new ActionListener () {
                public void actionPerformed (ActionEvent event) {
                    String fn = PopupDialog.askFilename 
                        (GUI.this,
                         "Choose Keystore File", 
                         keystoreFile.getText (), 
                         true);
                    if (fn != null)
                        keystoreFile.setText (fn);
                }
            });

            me.add (editKeys = new Button ("Manage Trusted Keys..."),
                    Constrain.labelLike (1, 2));
            me.add (clearKeys = new Button ("Regenerate All Keys"),
                    Constrain.labelLike (2, 2));
            me.add (encryptionRequired = 
                    new Checkbox ("Require encryption on remote connections",
                                  Encryption.isRequired ()),
                    Constrain.labelLike (1, 4, 2));
            me.add (encryptionRequiredLocally = 
                    new Checkbox ("Require encryption on local connections",
                                  Encryption.isRequiredLocally ()),
                    Constrain.labelLike (1, 5, 2));

            if (!Encryption.isSupported ()) {
                keystoreFile.setEnabled (false);
                encryptionRequired.setEnabled (false);
                encryptionRequiredLocally.setEnabled (false);
                editKeys.setEnabled (false);
                clearKeys.setEnabled (false);

                me.add (new MultiLineLabel 
                        ("Encryption is supported only in Java 1.2 or later with JSSE installed.\n"
                         + "See the Remote Clip web page for more details."),
                    Constrain.labelLike (1, 5));
            }

            editKeys.addActionListener (new ActionListener () {
                    public void actionPerformed (ActionEvent event) {
                        new KeystoreDialog ().show ();
                    }
                });

            clearKeys.addActionListener (new ActionListener () {
                    public void actionPerformed (ActionEvent event) {
                        regenerateKeys ();
                    }
                });

            KeyListener keyListener = new KeyAdapter () {
                public void keyPressed (KeyEvent event) {
                    switch (event.getKeyCode ()) {
                        case KeyEvent.VK_ENTER:
                        ok ();
                        break;
                        case KeyEvent.VK_ESCAPE:
                        cancel ();
                        break;
                    }
                }
            };
            OptionsDialog.this.addKeyListener (keyListener);
            port.addKeyListener (keyListener);
            encryptionRequired.addKeyListener (keyListener);
            encryptionRequiredLocally.addKeyListener (keyListener);
            
            OptionsDialog.this.pack ();
        }

        public void ok () {
            super.ok ();

            boolean newEncryptionRequired = encryptionRequired.getState ();
            boolean newEncryptionRequiredLocally = encryptionRequiredLocally.getState ();
            if (Encryption.isRequired () != newEncryptionRequired
                || Encryption.isRequiredLocally () != newEncryptionRequiredLocally) {
                Encryption.setRequired (newEncryptionRequired);
                Encryption.setRequiredLocally (newEncryptionRequiredLocally);

                if (!newEncryptionRequired || !newEncryptionRequiredLocally) {
                    try {
                        protocol.getProtocol (Locator.RMI).listen ();
                    } catch (Exception e) {
                        Global.report (e);
                        popupError (e.toString ());
                    }
                }
            }

            String newPort = port.getText ();
            if (!newPort.equals (originalPort)) {
                Global.setLocalPort (Integer.parseInt (newPort));
                locator = new Locator (locator.getProtocol (), 
                                       Global.getLocalHost (), 
                                       Global.getLocalPort ());
                clipboard.setLocator (locator);
                groupChanged ();

                Global.getConfig ().put ("Port", newPort);
                Global.getConfig ().save ();

                protocol.stopListening ();
                try {
                    protocol.listen ();
                } catch (Exception e) {
                    Global.report (e);
                    popupError ("Can't listen for connections on "
                                + newPort + ":\n"
                                + e.toString ());
                }
            }

            String newKeystoreFile = keystoreFile.getText ();
            if (!newKeystoreFile.equals (originalKeystoreFile)) {
                Global.getConfig ().put ("Keystore", newKeystoreFile);
                Global.getConfig ().save ();

                File f = new File (newKeystoreFile);
                Encryption.setKeystoreFile (f);

                try {
                    setHourglass (true);
                    reportStatus ("Loading keystore...");
                    Encryption.getInstance ().reloadKeys ();
                    setHourglass (false);
                    reportStatus ("Done.");
                } catch (Exception e) {
                    Global.report (e);
                    popupError (e.toString ());
                }
            }
        }
    }

    //
    // Regenerate keys
    //

    void regenerateKeys () {
        try {
            setHourglass (true);
            reportStatus ("Regenerating keys...");
            Encryption.getInstance ().regenerateKeys ();
            setHourglass (false);
            reportStatus ("Done.");
            
            PopupDialog.warn (this, 
                              "Regenerated Encryption Keys", 
                              "New encryption keys have been generated for this host.\n"
                              +"\n"
                              +"All trust relationships between this host and other hosts\n"
                              +"are now invalid, so you will have to authorize access\n"
                              +"when formerly-trusted hosts try to connect again.\n"
                              );
        } catch (Exception e) {
            Global.report (e);
            popupError (e.toString ());
        }
    }


    // Keystore dialog box
    
    class KeystoreDialog extends OkCancelDialog {
        List certs;             // certificates in keystore
        Encryption.Cert[] certArray;
        Button removeButton;    // removes a certificate

        public KeystoreDialog () {
            super (Win.findFrame (GUI.this), 
                   "Trusted Keys");            

            Panel mainPanel = getControlsPane ();

            mainPanel.setLayout (new GridBagLayout ());

            Constrain.add (mainPanel, 
                           new Label ("Trusted Keys:"), 
                           Constrain.fieldLike (0, 0, 2));
            
            Constrain.add (mainPanel, certs = new List (5), 
                           Constrain.areaLike (0, 1));

            certArray = Encryption.getInstance ().getCertificates ();
            for (int i = 0; i < certArray.length; ++i)
                    certs.add (certArray[i].getName () + " (" 
                               + certArray[i].getFingerprint ()
                               + ")");
            certs.setMultipleMode (true);
            certs.addItemListener (new ItemListener () {
                public void itemStateChanged (ItemEvent event) {
                    int[] indexes = certs.getSelectedIndexes ();
                    removeButton.setEnabled (indexes != null
                                             && indexes.length > 0);
                }
            });


            Constrain.add (mainPanel, removeButton = new Button ("Remove"), 
                           Constrain.labelLike (1, 1));
            removeButton.setEnabled (false);
            removeButton.addActionListener (new ActionListener () {
                public void actionPerformed (ActionEvent event) {
                    int[] indexes = certs.getSelectedIndexes ();
                    if (indexes != null)
                        for (int i = indexes.length-1; i >= 0; --i) {
                            certArray[i].delete ();
                            System.arraycopy (certArray, i+1,
                                              certArray, i,
                                              certArray.length - i - 1);
                            certs.remove (indexes[i]);
                        }

                    removeButton.setEnabled (false);
                }
            });
            
            KeystoreDialog.this.pack ();
            PopupDialog.centerWindow (this, GUI.this);
        }
        
        public void ok () {
            super.ok ();
            KeystoreDialog.this.setVisible (false);
        }
        
        public void cancel () {
            super.cancel ();
        }
        
    };

}

