/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.awt.datatransfer.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

public class JavaClipboard extends SynchronizedClipboard implements ClipboardOwner {

    protected boolean acquiring = false;  
    // true iff we're in the midst of calling setContents() on the
    // system clipboard.  Used by lostOwnership() to tell whether 
    // system clipboard ownership is being lost to another application or
    // just to another setContents() made by this object.

    Button delayButton; 
    // invisible button used to call reallyLostOwnership() until after system
    // clipboard actually changes

    public JavaClipboard () {
        delayButton = new Button ("");
        delayButton.setVisible (false);
        delayButton.addActionListener (new ActionListener () {
            public void actionPerformed (ActionEvent event) {
                reallyLostOwnership ();
            }
        });

        seizeSystemClipboard (); 
    }

    //
    // Transferable that fetches data from system clipboard
    //

    Transferable systemClipboardContents = new Transferable () {
        public Object getTransferData (DataFlavor flavor) throws UnsupportedFlavorException, IOException {
            try {
                enter ();

                // Seize the system clipboard, because we're losing
                // exclusive possession of the synchronized clipboard.
                // In order to detect the next local copy operation, we need
                // to have ownership of the system clipboard.
                return seizeSystemClipboard ()
                    .getTransferData (flavor);
            } finally {
                leave ();
            }
        }

        public DataFlavor[] getTransferDataFlavors () {
            try {
                enter ();
                Transferable t = getSystemClipboard ()
                    .getContents (null);
                Global.debug ("system clipboard is " + t);
                return t
                    .getTransferDataFlavors ();
            } finally {
                leave ();
            }
        }

        public boolean isDataFlavorSupported (DataFlavor flavor) {
            try {
                enter ();
                return getSystemClipboard ()
                    .getContents (null)
                    .isDataFlavorSupported (flavor);
            } finally {
                leave ();
            }
        }

        // Check for infinite recursion.  This isn't supposed to
        // happen, but programming error could cause system clipboard
        // and synchronized clipboard to reference each other.
        boolean recursed = false;
        void enter () {
            if (recursed)
                throw new RuntimeException ("infinite recursion");
            recursed = true;
        }
        void leave () {
            recursed = false;
        }
    };

    //
    // Transferable that fetches data from the synchronized clipboard
    //

    Transferable syncClipboardContents = 
        new Transferable () {
        public Object getTransferData (DataFlavor flavor) 
            throws UnsupportedFlavorException, IOException {
            return getContents (null).getTransferData (flavor);
        }
        
        public DataFlavor[] getTransferDataFlavors () {
            return getContents (null).getTransferDataFlavors ();
        }
        
        public boolean isDataFlavorSupported (DataFlavor flavor) {
            return getContents (null).isDataFlavorSupported (flavor);
        }
    };
    
    // Grab the current system clipboard contents, then
    // assert ownership of the system clipboard and put the
    // very same data back on it.  We have to seize ownership
    // like this when we want to detect copy operations in other
    // applications, since only the clipboard owner is notified when the 
    // clipboard changes.
    protected Transferable seizeSystemClipboard () {
        Global.debug ("seizing ownership of system clipboard");
        Clipboard systemClipboard = getSystemClipboard ();
        Transferable contents = systemClipboard.getContents (null);

        try {
            acquiring = true;
            systemClipboard.setContents (contents, this);
        } finally {
            acquiring = false;
        }

        return contents;
    }


    // When synchronized clipboard is set to an external
    // contents (like a remote clipboard), we assert ownership
    // of the system clipboard and put syncClipboardContents on it. 
    protected void acquireSystemClipboard (String mimeType) {
        Global.debug ("putting remote contents on system clipboard");
        try {
            acquiring = true;
            getSystemClipboard ().setContents (syncClipboardContents, this);
        } finally {
            acquiring = false;
        }
    }

    static Clipboard getSystemClipboard () {
        return Toolkit.getDefaultToolkit ().getSystemClipboard ();
    }


    // When another application changes the system clipboard,
    // we get notified by through the ClipboardOwner interface.
    // Deal with it by asserting ownership of the synchronized
    // clipboard and putting systemClipboardContents on it.
    public void lostOwnership (Clipboard clipboard, Transferable contents) {
        if (!acquiring) {
            Global.debug ("losing ownership of system clipboard");

            // We're going to lose ownership, but the system clipboard
            // hasn't changed yet.
            // Use the AWT event queue to call reallyLostOwnership () 
            // in a later iteration of the event-handling loop, after
            // our caller has actually changed the system clipboard to
            // its new contents.
            postActionEvent (delayButton);
        }
    }

    public void reallyLostOwnership () {
        Global.debug ("lost ownership of system clipboard");
        setContents (systemClipboardContents, SYSTEM_OWNER);
    }

    void postActionEvent (Component c) {
        AWTEvent event = new ActionEvent (c, 
                                          ActionEvent.ACTION_PERFORMED,
                                          null);
        Toolkit toolkit = Toolkit.getDefaultToolkit ();
        EventQueue queue = toolkit.getSystemEventQueue ();
        queue.postEvent (event);
    }
}
