/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.io.*;
import java.net.*;
import java.util.*;

public class Locator implements Serializable { 
    static final long serialVersionUID = 8775070821837646192L; 

    int protocol;           // One of RMI, ...
    String host;            // Hostname
    String ipaddr;          // IP address (n.n.n.n), possibly null
    transient InetAddress addr;       // IP address
    int port;               // Port number

    public static final int SECURE_RMI = 0;
    public static final int RMI = 1;
    public static final int SERIAL = 2;
    public static final int NPROTOCOLS = 3;

    public static final String[] PROTOCOL_PREFIX = { 
        "rmis:",
        "rmi:",
        "serial:",
    };

    public static final int DEFAULT_PROTOCOL = 
        Encryption.isSupported () ? SECURE_RMI : RMI;
    public static final int DEFAULT_PORT = 9494;

    public Locator (int protocol, InetAddress addr, int port) {
        this.protocol = protocol;
        this.host = addr.getHostName ();
        this.addr = addr;
        this.port = port;
    }

    public Locator (String spec) {
        // Parse and strip off protocol, if any
        protocol = DEFAULT_PROTOCOL;
        for (int i = 0; i < PROTOCOL_PREFIX.length; ++i) {
            if (spec.startsWith (PROTOCOL_PREFIX[i])) {
                protocol = i;
                spec = spec.substring (PROTOCOL_PREFIX[i].length());
                break;
            }
        }

        if (protocol == SERIAL) {
            // everything else is the port name
            host = spec;
            return;
        }

        if (spec.startsWith ("//"))
            spec = spec.substring (2);

        // Parse and strip off port number, if any
        port = DEFAULT_PORT;
        int colon = spec.indexOf (':');
        if (colon != -1) {
            port = Integer.parseInt (spec.substring (colon+1));
            spec = spec.substring (0, colon);
        }


        // Parse and strip off IP address (to avoid DNS lookups wherever
        // possible)
        int slash = spec.indexOf ('/');
        if (slash != -1) {
            ipaddr = spec.substring (slash+1);
            spec = spec.substring (0, slash);
        }

        // Rest of spec is hostname
        if (spec.length () > 0)
            host = spec;
    }

    public int getProtocol () {
        return protocol;
    }

    public String getHostName () {
        return host;
    }

    public InetAddress getAddress () throws IOException {
        if (addr == null)
            addr = InetAddress.getByName ((ipaddr != null) ? ipaddr : host);
        return addr;
    }

    public int getPort () {
        return port;
    }

    public String toString () {
        switch (protocol) {
        case RMI:
        case SECURE_RMI:
            return PROTOCOL_PREFIX[protocol]
                + getHostNameAndAddress ()
                + ":"
                + port;
        case SERIAL:
            return PROTOCOL_PREFIX[protocol]
                + host;
        default:
            return "unknown";
        }
    }

    public String getHostNameAndAddress () {
        String h = host;

        if (addr != null)
            return h + "/" + addr.getHostAddress ();
        else if (ipaddr != null)
            return h + "/" + ipaddr;
        else
            return h;
    }

    public String getHumanName () {
        switch (protocol) {
        case RMI:
            return getRMIName () + " (unencrypted)";
        case SECURE_RMI:
            return getRMIName ();
        case SERIAL:
            return host + " (serial port)";
        default:
            return toString ();
        }

    }

    String getRMIName () {
        StringBuffer result = new StringBuffer ();

        if (host != null)
            result.append (host.toLowerCase ());
        else if (addr != null)
            result.append (addr.getHostAddress ());
        else if (ipaddr != null)
            result.append (ipaddr);

        if (port != DEFAULT_PORT) {
            result.append (':');
            result.append (port);
        }

        return result.toString ();
    }

    public boolean equals (Object obj) {
        if (! (obj instanceof Locator))
            return false;

        Locator loc = (Locator)obj;

        return 
            protocol == loc.protocol
            && port == loc.port
            && ((addr != null && loc.addr != null && addr.equals (loc.addr))
                || (ipaddr != null && (ipaddr.equals (loc.ipaddr)
                                       || ipaddr.equals (loc.host)))
                || (host != null && (host.equalsIgnoreCase (loc.host)
                                     || host.equals (loc.ipaddr))))
            ;
    }

}
