/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.awt.datatransfer.*;
import java.net.*;
import java.io.*;
import java.util.*;
import rcm.io.Line;

public class MetaProtocol implements Protocol {
    SynchronizedClipboard syncClipboard;

    static String[] protocolClasses = new String[] {
        "SecureRMIProtocol",
        "RMIProtocol",
        "SerialProtocol"
    };

    static String[] protocolHello = new String[] {
        "RCLIP/RMI-SSL 3.0",
        "RCLIP/RMI 3.0",
        "*** serial port ***"
    };

    public static final int MAXLINE = 100;

    Protocol[] protocols = new Protocol[protocolClasses.length];

    ListenThread listenThread;

    public MetaProtocol () {
    }
    
    public void setSynchronizedClipboard (SynchronizedClipboard syncClipboard) {
        this.syncClipboard = syncClipboard;

        for (int i = 0; i < protocols.length; ++i) {
            Protocol p = protocols[i];
            if (p != null)
                p.setSynchronizedClipboard (syncClipboard);
        }
    }

    //
    // listen for connections
    //

    public void listen () throws Exception {
        // tell subprotocols to start listening

        // secure RMI
        if (Encryption.isSupported ())
            getProtocol (Locator.SECURE_RMI).listen ();

        // insecure RMI
        if (!Encryption.isRequired ()
            || !Encryption.isRequiredLocally ()
            || !Encryption.isSupported ())
            getProtocol (Locator.RMI).listen ();

        // serial port
        if (Serial.isSupported ())
            getProtocol (Locator.SERIAL).listen ();

        // bind socket to port
        int port = Global.getLocalPort ();
        ServerSocket serverSocket = new ServerSocket (port);

        // start listening
        listenThread = new ListenThread (serverSocket);
        listenThread.start ();

        Global.debug ("now listening at port " + port);
    }


    public void stopListening () {
        if (listenThread != null) {
            Global.debug ("stopping listen thread");
            Thread t = listenThread;
            listenThread = null;
            t.interrupt ();
        }

        // tell subprotocols to stop listening
        getProtocol (Locator.RMI).stopListening ();
        if (Encryption.isSupported ())
            getProtocol (Locator.SECURE_RMI).stopListening ();
    }

    class ListenThread extends Thread {
        ServerSocket serverSocket;

        public ListenThread (ServerSocket serverSocket) {
            this.serverSocket = serverSocket;
        }
        public void run () {
            try {
                serverSocket.setSoTimeout (5000);
            } catch (IOException e) {
                Global.report (e);
            }

            while (listenThread == Thread.currentThread ()) {
                Socket socket = null;
                try {
                    socket = serverSocket.accept();
                    new ServiceThread (socket).start ();
                } catch (IOException e) {
                }
            }

            try {
                serverSocket.close ();
            } catch (IOException e) {
            }

            Global.debug ("old listening thread stopped");
        }
    }

    class ServiceThread extends Thread {
        Socket socket;
        public ServiceThread (Socket socket) {
            this.socket = socket;
        }
        public void run () {
            try {
                handleConnection (socket);
            } catch (Exception e) {
                Global.report (e);
            }

            try {
                socket.close ();
            } catch (IOException e2) {
            }
        }
    }

    public void handleConnection (Socket socket) throws Exception {
        String hello = Line.readLine (socket.getInputStream (), MAXLINE);

        for (int i = 0; i < protocols.length; ++i) {
            if (protocolHello[i].equals (hello)) {
                if (protocols[i] == null)
                    break;
                protocols[i].handleConnection (socket);
                return;
            }
        }
        
        Line.writeLine (socket.getOutputStream (), "201 protocol not supported");
    }

    //
    // forward other protocol operations
    //

    public void join (Locator locator) throws Exception {
        getProtocol (locator).join (locator);
    }
    
    public Clipboard open (Locator locator) throws Exception {
        return getProtocol (locator).open (locator);
    }

    public Protocol getProtocol (Locator locator) {
        return getProtocol (locator.getProtocol ());
    }

    public Protocol getProtocol (int p) {
        Protocol protocol;

        protocol = protocols[p];
        if (protocol != null)
            return protocol;

        String className = protocolClasses[p];
        try {
            protocol = 
                (Protocol) Class.forName (className).newInstance ();
        } catch (Throwable e) {
            Global.report (e);
            throw new RuntimeException (e.toString ());
        }

        if (syncClipboard != null)
            protocol.setSynchronizedClipboard (syncClipboard);
        protocols[p] = protocol;
        return protocol;
    }

    public static String getProtocolHello (int p) {
        return protocolHello[p];
    }

}
