/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.awt.datatransfer.*;
import java.io.*;
import java.net.*;
import java.rmi.*;
import java.rmi.server.*;
import java.rmi.registry.*;
import java.util.*;
import rcm.io.Line;

public class RMIProtocol implements Protocol {
    SynchronizedClipboard syncClipboard; // My clipboard 
    Locator locator;               // My hostname & port

    // remote clipboard service
    RMIClipboard remoteClipboard;
    RemoteStub remoteClipboardStub;

    // server
    volatile Thread listenThread;
    ServerSocket serverSocket;

    public RMIProtocol () throws Exception {
        if (Global.getDebug ().isEnabled ())
            RemoteServer.setLog (System.err);
    }

    public void setSynchronizedClipboard (SynchronizedClipboard syncClipboard) {
        this.syncClipboard = syncClipboard;
    }

    protected int getProtocolID () {
        return Locator.RMI;
    }

    //
    // creating remote clipboard
    //

    public void listen () throws Exception {
        locator = new Locator (getProtocolID (), 
                               Global.getLocalHost (),
                               Global.getLocalPort ());

        if (remoteClipboard == null) {
            // create and export remote clipboard
            remoteClipboard = new RMIClipboardImpl (this);
            remoteClipboardStub = export (remoteClipboard); 
            Global.debug ("exported clipboard as " + remoteClipboardStub);
        }
    }
    
    protected RemoteStub export (Remote remote) throws RemoteException {
        return (RemoteStub) UnicastRemoteObject.exportObject (remote);
    }


    protected FileList exportFileList (FileList list) throws IOException {
        return new RemoteFileList (list);
    }

    public void stopListening () {
    }

    public void handleConnection (Socket socket) throws Exception {
        OutputStream out = socket.getOutputStream ();

        if (!authorize (socket.getInetAddress ())) {
            Line.writeLine (out, "202 not authorized");
            return;
        }

        Line.writeLine (out, "100 OK");

        writeClipboard (out);
    }

    protected boolean authorize (InetAddress addr) {
        // special handling for localhost
        if (Global.isLocalHost (addr))
            return !Encryption.isRequiredLocally ();

        // otherwise it's a remote host
        if (Encryption.isRequired ())
            return false;

        String hostname = addr.getHostName ();
        Global.debug ("checking for authorization for " 
                                  + addr 
                                  + ", resolved hostname = " 
                                  + hostname);

        // give access automatically if IP address is already
        // found in clipboard group
        Locator[] group = syncClipboard.getLocators ();
        for (int i = 0; i < group.length; ++i) {
            try {
                if (group[i].getAddress ().equals (addr))
                    return true; // access permitted
            } catch (IOException e) {
            }
        }

        // ask user for manual authorization
        UI ui = Global.getUI ();
        String msg = 
            hostname
            + " is trying to make an unencrypted connection"
            + " to your clipboard.\n"
            + "Do you want to allow this access?";
        if (ui.authorize (msg))
            return true;

        ui.reportStatus ("Refused connection from " + hostname);
        return false;
    } 


    protected void writeClipboard (OutputStream out) throws Exception {
        ObjectOutputStream objout = new ObjectOutputStream (out);
        objout.writeObject (remoteClipboardStub);
        objout.flush ();
    }

    //
    // fetching a remote clipboard
    //

    protected RMIClipboard getClipboard (Locator remote) throws Exception {
        // get host and port
        String host = remote.getAddress().getHostAddress ();
        int port = remote.getPort ();

        // open socket
        Global.debug ("opening socket to " + remote);
        Socket socket = new Socket (host, port);

        try {
            makeRequest (socket);
            return readClipboard (socket);
        } finally {
            socket.close ();
        }
    }

    protected void makeRequest (Socket socket) throws Exception {
        Line.writeLine (socket.getOutputStream (),
                        MetaProtocol.getProtocolHello (getProtocolID ()));
        
        String response = Line.readLine (socket.getInputStream (), MetaProtocol.MAXLINE);
        if (!response.startsWith ("100"))
            throw new IOException (response);
    }

    protected RMIClipboard readClipboard (Socket socket) throws Exception {
        // read stub for remote clipboard
        ObjectInputStream in = 
            new ObjectInputStream (socket.getInputStream ());

        RMIClipboard clipboard = null;
        Exception error = null;
        try {
            Global.debug ("reading stub");
            clipboard = (RMIClipboard) in.readObject ();
            Global.debug ("stub = " + clipboard);
        } catch (Exception e) {
            Global.report (e);
            error = e;
        }

        if (error != null)
            throw error;
        return clipboard;
    }

    //
    // opening a remote clipboard
    //

    public Clipboard open (Locator remote) throws Exception {
        RMIClipboard clipboard = getClipboard (remote);
        return new RMIClipboardProxy (clipboard, this);
    }
    
    // 
    // joining a remote clipboard group
    //

    public void join (Locator remote) throws Exception {
        try {
            RMIClipboard rc = getClipboard (remote);
            Global.debug ("clipboard = " + rc);
            rc.connect (remoteClipboard);
            remoteClipboard.connect (rc);
        } catch (RemoteException e) {
            if (e.detail != null) {
                Global.report (e);
                throw (Exception)e.detail;
            } else
                throw e;
        }
    }


    static boolean isClientLocalHost () {
        String clientHostname;
        try {
            clientHostname = RemoteServer.getClientHost ();
        } catch (ServerNotActiveException e) {
            Global.report (e);
            throw new RuntimeException (e.toString ());
        }

        InetAddress clientAddr;
        try {
            clientAddr = InetAddress.getByName (clientHostname);
        } catch (java.net.UnknownHostException e) {
            Global.report (e);
            clientAddr = null;
        }

        return Global.isLocalHost (clientAddr);
    }

}

/**
 * Clipboard service, for accessing another host's remote clipboard interface.
 */
interface RMIClipboard extends Remote {
    public Locator getLocator () throws RemoteException;
    public String getMimeType () throws RemoteException;
    public ConstantTransferable getContents () throws RemoteException;
    public void setContents (RMIClipboard clipboard,
                             String mimeType,
                             RemoteOwner owner) throws RemoteException;
    public void setContents (ConstantTransferable contents) throws RemoteException;
    public void connect (RMIClipboard clipboard) throws RemoteException;
    public void disconnect (RMIClipboard clipboard) throws RemoteException;
    public void remove (RMIClipboard clipboard) throws RemoteException;
}

/**
 * Implementation of RMIClipboard.
 */
class RMIClipboardImpl extends RemoteObject implements RMIClipboard {
    RMIProtocol protocol;

    public RMIClipboardImpl (RMIProtocol protocol) throws RemoteException {
        this.protocol = protocol;
    }

    public Locator getLocator () throws RemoteException {
        return protocol.locator;
    }

    public String getMimeType () throws RemoteException {
        return
            Flavors.getMimeType (protocol.syncClipboard.getContents (protocol.locator));
    }

    public ConstantTransferable getContents () throws RemoteException {
        Transferable contents = protocol.syncClipboard.getContents (protocol.locator);
        ConstantTransferable t;
        if (contents instanceof ConstantTransferable)
            t = ((ConstantTransferable) contents);
        else
            t = new ConstantTransferable (contents);

        // convert FileList to RemoteFileList, if necessary
        if (t.isDataFlavorSupported (Flavors.filesFlavor)
            && !RMIProtocol.isClientLocalHost ()) {
            try {
                FileList list = 
                    (FileList) contents.getTransferData (Flavors.filesFlavor);
                FileList remoteList = protocol.exportFileList (list);
                if (remoteList != list)
                    t = new ConstantTransferable (remoteList);
            } catch (IOException e) {
                Global.report (e);
            } catch (UnsupportedFlavorException e) {
                Global.report (e);
            }
        }

        return t;
    }

    public void setContents (RMIClipboard clipboard,
                             String mimeType,
                             RemoteOwner owner) throws RemoteException {
        Transferable t = new RMITransferableProxy (clipboard, mimeType);
        protocol.syncClipboard.setContents (t, owner);
    }

    public void setContents (ConstantTransferable contents) throws RemoteException {
        protocol.syncClipboard.setContents (contents, null);
    }

    public void connect (RMIClipboard clipboard) throws RemoteException {
        try {
            Clipboard c = new RMIClipboardProxy (clipboard, protocol);
            protocol.syncClipboard.add (c);
        } catch (Throwable e) {
            Global.report (e);
            throw new RemoteException (e.toString ());
        }
    }

    public void disconnect (RMIClipboard clipboard) throws RemoteException {
        Clipboard c = new RMIClipboardProxy (clipboard, protocol);
        protocol.syncClipboard.disconnect (c);
    }

    public void remove (RMIClipboard clipboard) throws RemoteException {
        Clipboard c = new RMIClipboardProxy (clipboard, protocol);
        protocol.syncClipboard.remove (c);
    }
}

/**
 * Proxy for a remote clipboard which obeys Clipboard interface and
 * shields clients from network errors.
 */
class RMIClipboardProxy extends Clipboard implements Closable, Removable, Locatable {
    RMIClipboard clipboard;
    RMIProtocol protocol;
    Locator locator;

    public RMIClipboardProxy (RMIClipboard clipboard,
                              RMIProtocol protocol) throws RemoteException {
        super (null);
        this.clipboard = clipboard;
        this.protocol = protocol;
        this.locator = clipboard.getLocator ();
    }

    public Transferable getContents (Object requestor) {
        return new RMITransferableProxy (clipboard, null);
    }

    public void setContents (Transferable contents, ClipboardOwner owner) {
        try {
            if (contents instanceof ConstantTransferable) {
                // pass contents by value.  Used by RemoteCopy, since it
                // won't be around long enough to serve up the actual
                // data when it's requested.
                Global.debug ("notifying " + locator);
                clipboard.setContents ((ConstantTransferable) contents);
            } else {
                RemoteOwner rowner;
                if (owner instanceof RemoteOwner) {
                    rowner = (RemoteOwner) owner;
                    if (rowner.contains (locator)) {
                        // remote clipboard is on the path of this
                        // notification, so it already knows.
                        Global.debug ("skipping " + locator);
                        return;
                    }

                    rowner = new RemoteOwner (rowner, protocol.locator);
                } else
                    rowner = new RemoteOwner (protocol.locator);
                
                // pass contents by reference
                Global.debug ("notifying " + locator);
                clipboard.setContents (protocol.remoteClipboard,
                                       Flavors.getMimeType (contents),
                                       rowner);
            }
        } catch (RemoteException e) {
            Global.report (e);
            protocol.syncClipboard.disconnect (this);
        }
    }

    public void close () {
        try {
            Global.debug ("disconnecting from " + getName ());
            clipboard.disconnect (protocol.remoteClipboard);
        } catch (RemoteException e) {
            Global.report (e);
        }
    }

    public void remove () {
        try {
            Global.debug ("removing from " + getName ());
            clipboard.remove (protocol.remoteClipboard);
        } catch (RemoteException e) {
            Global.report (e);
        }
    }

    public String getName () {
        return getLocator ().getHumanName ();
    }

    public Locator getLocator () {
        return locator;
    }

    public boolean equals (Object obj) {
        return obj instanceof RMIClipboardProxy
            && ((RMIClipboardProxy)obj).locator.equals (locator);
    }
}

/**
 * Proxy for a remote clipboard's contents which obeys the Transferable
 * interface and shields clients from network errors.
 */
class RMITransferableProxy implements Transferable,Locatable {
    RMIClipboard clipboard;
    String mimeType;  // caches clipboard's current mime type
    Locator locator;  // caches clipboard.getLocator ()

    public RMITransferableProxy (RMIClipboard clipboard, String mimeType) {
        this.clipboard = clipboard;
        this.mimeType = mimeType;
    }

    public Object getTransferData (DataFlavor flavor) throws UnsupportedFlavorException, IOException {
        try {
            if (!Flavors.toMimeType (flavor).equals (getMimeType ()))
                throw new UnsupportedFlavorException (flavor);

            return clipboard.getContents ().getTransferData (flavor);
        } catch (RemoteException e) {
            Global.report (e);
            throw new IOException (e.toString ());
        }
    }

    public DataFlavor[] getTransferDataFlavors () {
        return Flavors.toDataFlavors (getMimeType ());
    }

    public boolean isDataFlavorSupported (DataFlavor flavor) {
        return Flavors.isDataFlavorSupported (flavor, 
                                              getTransferDataFlavors ());
    }

    public String getMimeType () {
        if (mimeType == null) {
            try {
                mimeType = clipboard.getMimeType ();
            } catch (RemoteException e) {
                Global.report (e);
            }
        }
        return mimeType;
    }

    public Locator getLocator () {
        if (locator == null) {
            try {
                locator = clipboard.getLocator ();
            } catch (RemoteException e) {
                Global.report (e);
            }
        }
        return locator;
    }
}
