/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.awt.datatransfer.*;
import java.util.*;
import java.net.*;
import java.io.*;
import rcm.awt.*;
import rcm.util.*;

public abstract class RemoteClip {

    public static void main (String[] argv) throws Exception {
        // set defaults
        boolean showGUI = true;
        Vector remoteServers = new Vector ();
        String clipboardClass = "JavaClipboard";
        Locator locator = null;

        // parse command-line arguments
        for (int i=0; i<argv.length; ++i) {
            if (argv[i].startsWith ("-")) {
                String opt = argv[i].substring (1);
                if (opt.equals ("debug")) {
                    Global.setDebug (Debug.VERBOSE);
                } else if (opt.equalsIgnoreCase ("class")) {
                    if (++i >= argv.length)
                        usage ();
                    clipboardClass = argv[i];
                } else if (opt.equals ("host")) {
                    if (++i >= argv.length)
                        usage ();
                    Global.setLocalHost (argv[i]);
                } else if (opt.equals ("port")) {
                    if (++i >= argv.length)
                        usage ();
                    Global.setLocalPort (Integer.parseInt (argv[i]));
                } else if (opt.equals ("keystore")) {
                    if (++i >= argv.length)
                        usage ();
                    Encryption.setKeystoreFile (new File (argv[i]));
                } else if (opt.equals ("nogui")) {
                    showGUI = false;
                } else if (opt.equals ("tty")) {
                    showGUI = false;
                    clipboardClass = "TTYClipboard";
                } else if (opt.equals ("v") || opt.equals ("version")) {
                    System.err.println ("Remote Clip v" + Global.VERSION);
                    System.exit (-1);
                } else if (opt.equals ("add")) {
                    if (++i >= argv.length)
                        usage ();
                    remoteServers.addElement (new Locator (argv[i]));
                } else
                    usage ();
            } else
                usage ();
        }

        // load clipboard class
        SynchronizedClipboard clipboard = makeClipboard (clipboardClass);

        // tell clipboard where it is
        locator = new Locator (Locator.DEFAULT_PROTOCOL, 
                               Global.getLocalHost (),
                               Global.getLocalPort ());
        clipboard.setLocator (locator);

        // make network protocol
        Global.debug ("making clipboard protocol");
        MetaProtocol protocol = new MetaProtocol ();
        protocol.setSynchronizedClipboard (clipboard);
        clipboard.setProtocol (protocol);

        // make the UI appear
        if (showGUI)
            Global.setUI (new GUI (clipboard, locator, protocol));

        UI ui = Global.getUI ();
        ui.reportStatus ("Loading. Please wait...");
        ui.setHourglass (true);

        // start listening
        Global.debug ("listening for remote connections");
        try {
            protocol.listen ();
        } catch (Exception ex) {
            Global.report (ex);
            ui.popupError ("Can't listen for remote connections.\n" 
                           + "Remote Clip will exit.");
            System.exit (-1);
        }


        // restore saved clipboard group
        if (!remoteServers.isEmpty ()) {
            Global.debug ("joining clipboard group from command line");
            // join hosts specified on command line
            for (Enumeration e = remoteServers.elements (); e.hasMoreElements (); )
                clipboard.add ((Locator) e.nextElement ());
        } else {
            Global.debug ("restoring saved clipboard group");
            clipboard.restoreGroup ();
        }

        ui.setHourglass (false);
        ui.reportStatus ("Ready.");
    }

    static SynchronizedClipboard makeClipboard (String clipboardClass) throws Exception {
        SynchronizedClipboard clipboard;

        Global.debug ("making local clipboard " + clipboardClass);

        int p = clipboardClass.indexOf ('-');
        if (p != -1) {
            // class name has form <java class>-<native library>
            String className = clipboardClass.substring (0, p);

            // load Java class
            Class cls = Class.forName (className);

            // create clipboard, passing it the library name
            clipboard = (SynchronizedClipboard)
                cls.getConstructor (new Class[] { String.class })
                .newInstance (new Object[] { clipboardClass });
        } else {
            // load Java class
            Class cls = Class.forName (clipboardClass);

            // create the clipboard with null constructor
            clipboard = (SynchronizedClipboard)
                cls.getConstructor (new Class[] { })
                .newInstance (new Object[] { });
        }

        return clipboard;
    }

    private static void usage () {
        System.err.println 
            (
             "Usage: rclip [options]\n"
             + "\n"
             + "    -add <clipboard>\n"
             + "       Add a host to the clipboard group.\n"
             + "       Any number of -add options can be listed.\n"
             + "       <clipboard> is one of:\n"
             + "\n"
             + "       rmi:<hostname>[:<port>]\n"
             + "              unencrypted RMI on <hostname>/<port>\n"
             + "\n"
             + (Encryption.isSupported ()
                 ? "       rmis:<hostname>[:<port>]\n"
                 + "              encrypted RMI on <hostname>/<port>\n"
                 + "\n"
                 : "")
             + (Serial.isSupported ()
                 ? "       serial:<serial port>\n"
                 + "              Palm Pilot on <serial port>\n"
                 + "\n"
                 : "")
             + "    -debug            show debugging output\n"
             + "    -class <class>    use clipboard class <class>\n"
             + "    -host <host>      use <host> as localhost name\n"
             + "    -port <port>      listen for connections on <port>\n"
             + "    -nogui            don't show window\n"
             + "    -tty              don't use X clipboard or show window\n"
             + "    -v                display version number\n"
             );
        System.exit (-1);
    }

}
