/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.zip.*;
import rcm.io.TempFile;
import rcm.progress.*;

public class RemoteFileList extends FileList {
    static final long serialVersionUID = 7696590927778897476L;

    String remoteHost;
    int remotePort;
    
    // server state (not serialized)
    transient ServerSocket server;
    transient RemoteFileList proxy;
    
    public RemoteFileList () {
    }

    public RemoteFileList (FileList fl) throws IOException {
        files = fl.files;
        names = fl.names;

        if (fl instanceof RemoteFileList)
            proxy = (RemoteFileList) fl;

        // Launch background server
        server = createServerSocket (0);
        new ServerThread ().start ();

        remoteHost = Global.getLocalHost ().getHostAddress ();
        remotePort = server.getLocalPort ();
    }

    //
    // creating sockets
    //

    protected Socket createSocket (String host,
                                   int port) throws IOException { 
        return new Socket (host, port);
    }
            
    protected ServerSocket createServerSocket (int port) throws IOException { 
        return new ServerSocket (0);
    }

    //
    // server
    //
    class ServerThread extends Thread {
        public void run () {
            try {
                // timeout if 4 minutes pass without a connection
                server.setSoTimeout (240000);
                try {
                    Socket client = server.accept ();
                    try {
                        OutputStream out = client.getOutputStream ();
                        if (proxy != null)
                            forward (out);
                        else
                            send (out);
                    } finally {
                        client.close ();
                    }
                } finally {
                    server.close ();
                }
            } catch (IOException e) {
                Global.report (e);
            }
        }
    }

    private void forward (OutputStream out) throws IOException {
        Global.debug ("forwarding remote files");

        Socket sock = proxy.openConnection ();
        InputStream in = sock.getInputStream ();

        byte[] buffer = new byte[BUFFER_SIZE];
        int n;
        while ((n = in.read (buffer, 0, buffer.length)) != -1)
            out.write (buffer, 0, n);

        sock.close ();
    }

    private void send (OutputStream out) throws IOException {
        byte[] buffer = new byte[BUFFER_SIZE];
        int n;

        ZipOutputStream zout = new ZipOutputStream (out);
        zout.setMethod (ZipOutputStream.STORED);
        
        for (Enumeration e = files.elements (); e.hasMoreElements ();)
            sendFile (zout, (File)e.nextElement (), "", buffer);
    }

    private void sendFile (ZipOutputStream zout, File f, String prefix, byte[] buffer) throws IOException {
        String name = prefix + f.getName ();
        
        if (f.isDirectory ()) {
            // Mark it as a directory for the ZipFile
            name += "/";
            
            // Create an entry for the directory
            ZipEntry entry = new ZipEntry (name);
            entry.setSize (0);
            entry.setCrc (0);
            
            // Add the entry to the ZIP archive
            zout.putNextEntry (entry);            
            
            // Add each file in directory recursively
            String[] subfiles = f.list ();
            for (int i = 0; i < subfiles.length; ++i)
                sendFile (zout, new File (f, subfiles[i]), name, buffer);
        }
        else {
            // Create an entry for this file
            ZipEntry entry = new ZipEntry (name);
            entry.setSize (f.length ());

            // compute the file's CRC
            Checksum crc = new CRC32 ();
            CheckedInputStream checkedInput = new CheckedInputStream (new FileInputStream (f), crc);
            while (checkedInput.read (buffer) != -1)
                ;
            checkedInput.close ();
            entry.setCrc (crc.getValue ());

            // Add the entry to the ZIP archive
            zout.putNextEntry (entry);
            
            // Copy the file's data into the ZIP archive
            InputStream in = new FileInputStream (f);
            int n;
            while ((n = in.read (buffer)) != -1)
                zout.write (buffer, 0, n);
            in.close ();
        }
    }


    //
    // client
    //

    private Socket openConnection () throws IOException {
        return createSocket (remoteHost, remotePort);
    } 

    public void copy (String directory, 
                      ProgressListener listener) throws IOException {
        byte[] buffer = new byte[BUFFER_SIZE];

        BasicProgressGenerator gen = new BasicProgressGenerator (this);
        if (listener != null)
            gen.addProgressListener (listener);

        // Make sure dir has a terminating separator
        if (directory == null)
            directory = ".";
        else if (directory.length () > 0
            && !(directory.endsWith ("/") ||
                 directory.endsWith (File.separator)))
            directory += File.separator;

        Socket sock = openConnection ();
        InputStream in = sock.getInputStream ();
        ZipInputStream zin = new ZipInputStream (in);

        try {
            ZipEntry entry;
            while ((entry = zin.getNextEntry ()) != null) {
                String name = entry.getName ();

                File target = new File (directory + entry.getName ());

                if (entry.isDirectory ()) {
                    gen.fireStartProgress (target.toString (), 1);
                    target.mkdir ();
                    gen.fireDoneProgress ();
                } else {
                    // Save the input stream to disk
                    OutputStream out = new FileOutputStream (target);
                    int length = (int)entry.getSize ();
                    int n;
                    gen.fireStartProgress (target.toString (), 
                                           (length + buffer.length - 1)
                                            / buffer.length);
                    while (length > 0 && 
                           (n = zin.read (buffer, 0, 
                                          Math.min (length, buffer.length))) 
                           != -1) {
                        out.write (buffer, 0, n);
                        length -= n;
                        gen.fireAddProgress (1);
                    }
                    
                    out.close ();
                    gen.fireDoneProgress ();
                }
                zin.closeEntry ();
            }
        } finally {
            zin.close ();
            sock.close ();
        }
    }
    

    public FileList makeLocal () throws IOException {
        File tempDir = new File (TempFile.getTempDirectory ());
        copy (tempDir.toString ());

        FileList fl = new FileList ();
        String[] names = getNames ();
        for (int i = 0; i < names.length; ++i)
            fl.add (new File (tempDir, names[i]));
        return fl;
    }
}

