/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.awt.datatransfer.*;
import java.util.*;
import java.net.*;
import java.io.*;
import rcm.awt.PopupDialog;
import rcm.util.*;

public abstract class RemotePaste {

    public static void main (String[] argv) throws Exception {
        // set defaults 
        String directory = null;
        Locator locator = null;

        // process command-line arguments
        for (int i=0; i<argv.length; ++i) {
            if (argv[i].startsWith ("-")) {
                String opt = argv[i].substring (1);
                if (opt.equals ("debug")) {
                    Global.setDebug (Debug.VERBOSE);
                } else if (opt.equals ("clipboard")) {
                    if (++i >= argv.length)
                        usage ();
                    locator = new Locator (argv[i]);
                } else if (opt.equals ("keystore")) {
                    if (++i >= argv.length)
                        usage ();
                    Encryption.setKeystoreFile (new File (argv[i]));
                } else if (opt.equals ("d")) {
                    if (++i >= argv.length)
                        usage ();
                    directory = argv[i];
                } else if (opt.equals ("v") || opt.equals ("version")) {
                    System.err.println ("Remote Clip v" + Global.VERSION);
                    System.exit (-1);
                } else
                    usage ();
            } else
                usage ();
        }

        // create locator
        if (locator == null) {
            String protocol = 
                (Encryption.isSupported () 
                 && Encryption.isRequiredLocally ())
                ? "rmis:"
                : "rmi:";
            locator = new Locator (protocol + "localhost");
        }
            
        // Make sure directory has a terminating separator
        if (directory == null)
            directory = ".";
        else if (directory.length () > 0
            && !(directory.endsWith ("/") ||
                 directory.endsWith (File.separator)))
            directory += File.separator;

        // Create network protocol
        Protocol protocol = new MetaProtocol ();
          
        // Create proxy for local clipboard
        Clipboard clipboard;
        try {
            clipboard = protocol.open (locator);
        } catch (IOException e) {
            if (Global.getDebug ().isEnabled ())
                Global.report (e);
            System.err.println (e 
                                + "\nIs Remote Clip running on this machine? "
                                + "Start it by running 'rclip'.");
            return;
        }

        // Fetch clipboard contents and make sure it's a FileList
        try {
            Transferable contents = clipboard.getContents (null);
            FileList files = 
                (FileList) contents.getTransferData (Flavors.filesFlavor);
            files.copy (directory, 
                        new rcm.progress.StreamProgressBar (System.err));
        } catch (UnsupportedFlavorException e) {
            System.err.println ("No files found on clipboard");
        }

    }

    private static void usage () {
        System.err.println 
            (
             "Usage: rpaste [options]\n"
             + "\n"
             + "    -d <directory>      paste files into <directory> instead\n"
             + "                        of current directory\n"
             + "\n"
             + "    -clipboard <protocol>:<hostname>[:<port>]\n"
             + "              contact clipboard on <hostname>/<port>\n"
             + "              using <protocol>.  Choices for <protocol>:\n"
             + "                 rmi     unencrypted RMI\n"
             + "                 rmis    encrypted RMI\n"
             + "              e.g. -clipboard rmi:blowfish\n"
             + "\n"
             + "    -debug            show debugging output\n"
             + "    -v                display version number\n"
             );
        System.exit (-1);
    }
}
