/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.awt.datatransfer.*;
import java.io.*;
import java.net.*;
import java.rmi.*;
import java.rmi.server.*;
import java.rmi.registry.*;
import javax.net.ssl.*;
import java.util.*;
import rcm.util.Config;
import java.security.*;
import java.security.cert.*;
import java.security.cert.Certificate;
import com.sun.net.ssl.SSLContext;
import com.sun.net.ssl.KeyManagerFactory;
import com.sun.net.ssl.KeyManager;
import com.sun.net.ssl.X509KeyManager;
import com.sun.net.ssl.TrustManagerFactory;
import com.sun.net.ssl.X509TrustManager;
import com.sun.net.ssl.TrustManager;
import rcm.util.Str;
import rcm.io.Line;

public class SSLEncryption extends Encryption {
    static final String LOCALHOST_ALIAS = "localhost";
    static final String DEFAULT_PASSPHRASE = "RemoteClip";
    static final char[] DEFAULT_PASSPHRASE_CHARS = 
        DEFAULT_PASSPHRASE.toCharArray ();

    KeyStore keystore;
    CertificateFactory cf;

    public SSLEncryption () {
        System.setProperty ("com.sun.net.ssl.dhKeyExchangeFix", "true");
        Security.addProvider (new com.sun.net.ssl.internal.ssl.Provider ());

        try {
            reloadKeys ();

            cf = CertificateFactory.getInstance ("X.509");

        } catch (Exception e) {
            Global.report (e);
        }

    }

    //
    // creating an SSLContext
    //

    public SSLContext makeSSLContext (boolean askUser) throws Exception {
        SSLContext sslContext = SSLContext.getInstance ("TLSv1");
        sslContext.init (new KeyManager[] {new MyKeyManager () },
                         new TrustManager[] { new MyTrustManager (askUser)},
                         null);
        return sslContext;
    }


    //
    // clearing encryption keys
    //

    public void regenerateKeys () throws Exception {
        makeKeyStore (Encryption.getKeystoreFile ());
        updateFingerprint ();
    }

    public void reloadKeys () throws Exception {
        File keystoreFile = Encryption.getKeystoreFile ();
        if (keystoreFile.exists ())
            loadKeyStore (keystoreFile);
        else
            makeKeyStore (keystoreFile);
        
        updateFingerprint ();
    }

    void updateFingerprint () {
        try {
            Certificate cert = getMyCertificate ();
            String fingerprint = getFingerprintFromCertificate (cert);
            Global.getUI ().setFingerprint ("Fingerprint: " 
                                            + fingerprint);
        } catch (Exception e) {
            Global.report (e);
        }
    }


    //
    // enumerating keystore
    //

    public Cert[] getCertificates () {
        Vector v = new Vector ();

        try {
            for (Enumeration e = keystore.aliases ();
                 e.hasMoreElements (); ) {
                String alias = (String) e.nextElement ();
                if (keystore.isCertificateEntry (alias))
                    v.addElement (new SSLCert (alias));
            }
        } catch (KeyStoreException e) {
            Global.report (e);
        }

        Cert[] certs = new Cert[v.size ()];
        v.copyInto (certs);
        return certs;
    }

    class SSLCert implements Cert {
        String alias;
        
        SSLCert (String alias) {
            this.alias = alias;
        }

        public String getName () {
            return alias;
        }

        public String getFingerprint () {
            try {
                return getFingerprintFromCertificate (keystore.getCertificate (alias));
            } catch (Exception e) {
                Global.report (e);
                return "unknown (" + e.toString () + ")";
            }
        }

        public void delete () {
            try {
                keystore.deleteEntry (alias);
                saveKeyStore ();
            } catch (Exception e) {
                Global.report (e);
            }
        }
    }

    


    //
    // key management
    //

    void makeKeyStore (File file) throws Exception {
        if (file.exists ())
            file.delete ();

        String username = System.getProperty ("user.name");
        if (username == null || username.length () == 0)
            username = "RemoteClip";

        String hostname = Global.getLocalHost ().getHostName ();

        Global.debug ("creating keystore " + file);
        String[] command = {
            "keytool",
            "-genkey",
            "-keyalg",
            "DSA",
            "-keystore",
            file.toString (),
            "-storepass",
            DEFAULT_PASSPHRASE,
            "-alias",
            LOCALHOST_ALIAS,
            "-dname",
            "CN=" + username + "@" + hostname,
            "-keypass",
            DEFAULT_PASSPHRASE
        };
        Process process = Runtime.getRuntime ().exec (command);
        process.waitFor ();
        Global.debug ("created keystore " + file);

        try {
            // use chmod to make keystore read/write only by current user.
            // only works under Unix, of course, but it's better than nothing.
            Global.debug ("protecting keystore " + file);
            command = new String[] {
                "chmod",
                "600",
                file.toString (),
            };
            process = Runtime.getRuntime ().exec (command);
            process.waitFor ();
            Global.debug ("protected keystore " + file);
        } catch (IOException e) {
            // not fatal, since chmod only found on Unix
        }

        loadKeyStore (file);
    }

    void loadKeyStore (File file) throws Exception {
        Global.debug ("using keystore " + file);
        keystore = KeyStore.getInstance ("JKS");
        keystore.load (new FileInputStream (file),
                       DEFAULT_PASSPHRASE_CHARS);
    }

    void saveKeyStore () throws Exception {
        FileOutputStream fout = 
            new FileOutputStream (Encryption.getKeystoreFile ());
        keystore.store (fout, DEFAULT_PASSPHRASE_CHARS);
        fout.close ();
    }

    class MyKeyManager implements X509KeyManager {
        public String chooseClientAlias (String keyType, Principal[] issuers) {
            return LOCALHOST_ALIAS;
        }

        public String[] getClientAliases (String keyType, Principal[] issuers) {
            return new String[] { LOCALHOST_ALIAS };
        }

        public String chooseServerAlias (String keyType, Principal[] issuers) {
            return LOCALHOST_ALIAS;
        }

        public String[] getServerAliases (String keyType, Principal[] issuers) {
            return new String[] { LOCALHOST_ALIAS };
        }

        public X509Certificate[] getCertificateChain (String alias) {
            try {
                return new X509Certificate[] {
                    (X509Certificate) getMyCertificate ()
                        };
            } catch (Exception e) {
                throw new RuntimeException (e.toString ());
            }
        }

        public PrivateKey getPrivateKey (String alias) {
            try {
                return getMyPrivateKey ();
            } catch (Exception e) {
                throw new RuntimeException (e.toString ());
            }
        }

    }

    class MyTrustManager implements X509TrustManager {
        boolean askUser;

        public MyTrustManager (boolean askUser) {
            this.askUser = askUser;
        }

        public X509Certificate[] getAcceptedIssuers () {
            return new X509Certificate[0];
        }

        public boolean isClientTrusted (X509Certificate[] chain) {
            return isTrusted (chain[0], null, askUser);
        }

        public boolean isServerTrusted (X509Certificate[] chain) {
            return isTrusted (chain[0], null, askUser);
        }

    }

    public Certificate getMyCertificate () throws Exception {
        return keystore.getCertificate (LOCALHOST_ALIAS);
    }

    public PrivateKey getMyPrivateKey () throws Exception {
        return (PrivateKey) keystore.getKey (LOCALHOST_ALIAS, 
                                             DEFAULT_PASSPHRASE_CHARS);
    }

    public String getUserName (Certificate cert) {
        if (cert instanceof X509Certificate) {
            String dn = ((X509Certificate) cert).getSubjectDN ().getName ();
            String cn = Str.before (Str.after (dn, "CN="), ",");
            if (cn.length () > 0)
                return cn;
        }

        // default
        return "RemoteClip";
    }

    public String getFingerprintFromCertificate (Certificate cert) {
        try {
            MessageDigest md = MessageDigest.getInstance ("MD5");
            md.update (cert.getEncoded ());
            byte[] fingerprint = md.digest ();
            
            StringBuffer buf = new StringBuffer ();
            for (int i = 0; i < fingerprint.length; ++i) {
                if (i > 0)
                    buf.append (':');
                
                String hex = Integer.toHexString (fingerprint[i] & 0xFF);
                if (hex.length () < 2)
                    hex = "0" + hex;
                hex = hex.toUpperCase ();
                buf.append (hex);
            }
            return buf.toString ();
        } catch (Exception e) {
            Global.report (e);
            return "unknown";
        }
    }

    public void writeCertificate (OutputStream out, Certificate cert) throws Exception {
        byte[] encoded = cert.getEncoded ();
        Line.writeLine (out, String.valueOf (encoded.length));
        out.write (encoded);
    }

    public Certificate readCertificate (InputStream in) throws Exception {
        String lenString = Line.readLine (in, MetaProtocol.MAXLINE);
        int len = Integer.parseInt (lenString);

        DataInputStream din = new DataInputStream (in);

        byte[] buf = new byte[len];
        din.readFully (buf);

        ByteArrayInputStream certin = new ByteArrayInputStream (buf);
        return cf.generateCertificate (certin);
    }

    public boolean isTrusted (Certificate cert, InetAddress addr, boolean askUser) {
        try {
            // assert that certificate is self-signed
            cert.verify (cert.getPublicKey ());
            
            // check whether certificate is already in keystore 
            String alias = keystore.getCertificateAlias (cert);
            if (alias != null) {
                Global.debug ("already trusted as "
                              + alias);
                return true;
            }
            
            // ask user to authorize
            if (!askUser
                || !userAuthorized (cert, addr))
                return false;
        } catch (Exception e) {
            Global.report (e);
            return false;
        }

        try {
            // user said OK, so add the key to the keystore
            // so that we can make SSL connections to the server
            String alias = getFreshAlias (getUserName (cert));
            keystore.setCertificateEntry (alias, cert);
            Global.debug ("stored certificate as "
                          + alias);

            saveKeyStore ();

        } catch (Exception e) {
            Global.report (e);
        }

        return true;
    }

    String getFreshAlias (String base) throws KeyStoreException {
        int i = 0;
        String alias = base;
        do {
            if (!keystore.containsAlias (alias))
                return alias;
            alias = base + (i++);
        } while (true);
    }

    boolean userAuthorized (Certificate cert, InetAddress addr) {
        String user = getUserName (cert);
        String fingerprint = getFingerprintFromCertificate (cert);

        Global.debug ("checking for authorization for " 
                                  + user);
        UI ui = Global.getUI ();

        String msg = 
            user
            + ((addr != null)
               ? " (from " +  addr.getHostName () + ")"
               : "")
            + " is trying to make an encrypted connection"
            + " to your clipboard.\n"
            + "The fingerprint is " + fingerprint + ".\n"
            + "\n"
            + "Do you want to allow this access?";
        if (ui.authorize (msg))
            return true;

        ui.reportStatus ("Refused connection from " 
                         + user);
        return false;
    }

}
