/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.awt.datatransfer.*;
import java.io.*;
import java.net.*;
import java.rmi.*;
import java.rmi.server.*;
import java.rmi.registry.*;
import javax.net.ssl.*;
import java.util.*;
import rcm.util.Config;
import java.security.*;
import java.security.cert.*;
import java.security.cert.Certificate;
import com.sun.net.ssl.SSLContext;
import rcm.util.Str;
import rcm.io.Line;

public class SecureRMIProtocol extends RMIProtocol {

    SSLEncryption ssl;
    SSLContext sslContext;

    public SecureRMIProtocol () throws Exception {
        if (!Encryption.isSupported ())
            throw new RuntimeException ("encryption not supported");

        theInstance = this;
        ssl = (SSLEncryption) Encryption.getInstance ();
        sslContext = ssl.makeSSLContext (false);
    }

    protected int getProtocolID () {
        return Locator.SECURE_RMI;
    }

    //
    // singleton instance
    //

    static SecureRMIProtocol theInstance;

    static SecureRMIProtocol getInstance () {
        if (theInstance == null) {
            Global.debug ("creating new SecureRMIProtocol!");
            try {
                theInstance = new SecureRMIProtocol ();
            } catch (Exception e) {
                Global.report (e);
            }
        }
        return theInstance;
    }


    //
    // client requests
    //

    protected void makeRequest (Socket socket) throws Exception {
        InputStream in = socket.getInputStream ();
        OutputStream out = socket.getOutputStream ();

        // send hello
        Line.writeLine (out,
                        MetaProtocol.getProtocolHello (getProtocolID ()));
        
        // send my certificate
        Certificate myCert = ssl.getMyCertificate ();
        ssl.writeCertificate (out, myCert);

        // read response
        String response = Line.readLine (in, MetaProtocol.MAXLINE);
        if (!response.startsWith ("100"))
            throw new IOException (response);

        // read server's certificate
        Certificate cert = ssl.readCertificate (in);

        // check it 
        ssl.isTrusted (cert, socket.getInetAddress (), true);
    }

    //
    // server responses 
    //

    public void handleConnection (Socket socket) throws Exception {
        InputStream in = socket.getInputStream ();
        OutputStream out = socket.getOutputStream ();

        // read client's certificate
        Certificate cert = ssl.readCertificate (in);

        // check it 
        if (!ssl.isTrusted (cert, socket.getInetAddress (), true)) {
            Line.writeLine (out, "202 not authorized");
            return;
        }

        // send OK response
        Line.writeLine (out, "100 OK");

        // send my certificate
        Certificate myCert = ssl.getMyCertificate ();
        ssl.writeCertificate (out, myCert);

        // send clipboard stub
        writeClipboard (out);
    }


    //
    // exporting the clipboard
    //
    
    protected RemoteStub export (Remote remote) throws RemoteException {
        Global.debug ("securely exporting " + remote);
        return (RemoteStub) 
            UnicastRemoteObject.exportObject 
            (remote,
             0, 
             new RMISSLClientSocketFactory(),
             new RMISSLServerSocketFactory());
    }

    static class RMISSLClientSocketFactory
	implements RMIClientSocketFactory, Serializable {
        
        public Socket createSocket(String host, int port)
            throws IOException
	{
            return getInstance ().createSocket (host, port);
	}
    }

    static class RMISSLServerSocketFactory
	implements RMIServerSocketFactory, Serializable {
        
        public ServerSocket createServerSocket(int port)
            throws IOException
	{
            return getInstance ().createServerSocket (port);
	}
    }

    //
    // exporting file lists
    //

    protected FileList exportFileList (FileList list) throws IOException {
        return new SecureRemoteFileList (list);
    }

    //
    // creating sockets
    //

    protected Socket createSocket (String host,
                                int port) throws IOException { 
        SSLSocket socket = (SSLSocket)
            sslContext
            .getSocketFactory ()
            .createSocket(host, port);

        socket.setNeedClientAuth (true);

        Global.debug ("starting handshake");
        socket.startHandshake ();
        Global.debug ("handshake complete");
        
        SSLSession session = socket.getSession ();
        Global.debug ("cipher suite = " 
                                  + session.getCipherSuite ());
        
        return socket;
    }
            
    protected ServerSocket createServerSocket (int port) throws IOException { 
        SSLServerSocket socket = (SSLServerSocket)
            sslContext
            .getServerSocketFactory ()
            .createServerSocket (port);

        socket.setNeedClientAuth (true);

        return socket;
    }
}

