/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.awt.datatransfer.*;
import java.util.*;
import java.io.*;
import java.net.*;

public abstract class SynchronizedClipboard extends Clipboard implements Closable {
    Vector group = new Vector ();  // clipboard group, Vector of Clipboards

    Protocol protocol;        // network protocol
    Locator myLocator;          // locator for this host

    DataFlavor[] flavors = Flavors.STRING_FLAVORS; 
    // caches getContents ().getDataFlavors () so that we can tell when data
    // with new flavors is put on the clipboard

    boolean exclusive = false;
    // true iff we have exclusive possession of the clipboard contents, i.e.:
    // we're the clipboard owner, and no remote clipboard has asked for
    // the data yet.  Exclusive possession lets us change the clipboard data
    // repeatedly without notifying the whole group each time.

    
    
    public SynchronizedClipboard () {
        super ("");
    }

    public void close () {
        saveGroup ();

        Clipboard[] clipboards;
        synchronized (group) {
            clipboards = getClipboards ();
            group.removeAllElements ();
        }

        for (int i = 0; i < clipboards.length; ++i) {
            Clipboard c = clipboards[i];
            if (c instanceof Closable)
                ((Closable)c).close ();
        }
    }

    //
    // network protocol
    //

    public Protocol getProtocol () {
        return protocol;
    }

    public void setProtocol (Protocol protocol) {
        this.protocol = protocol;
    }

    public Locator getLocator () {
        return myLocator;
    }

    public void setLocator (Locator locator) {
        try {
            Global.debug ("my IP address is " + locator.getAddress ());
        } catch (IOException e) {
            Global.debug ("don't know my IP address!");
        }
        this.myLocator = locator;
    }



    //
    // Clipboard interface
    //

    public String getName () {
        return (myLocator != null) 
            ? (myLocator.getHumanName () + " (local host)")
            : "synchronized clipboard (local host)";
    }
 
    public void setContents (Transferable contents, ClipboardOwner owner) {
        Global.debug ("setContents from clipboard owner " + owner);

        DataFlavor[] newFlavors = contents.getTransferDataFlavors ();
        Global.debug ("clipboard is becoming " + Flavors.getName (newFlavors));

        // decide whether we have exclusive ownership of clipboard data,
        // and whether we need to notify other group members
        boolean notify;
        if (owner instanceof RemoteOwner) {
            Global.debug ("remote owner has exclusive ownership");
            exclusive = false;
            notify = true;
        } else if (exclusive && Flavors.areEqual (flavors, newFlavors)) {
            // we have exclusive possession of clipboard content,
            // and only the content has changed (not the flavors);
            // don't need to notify group
            Global.debug 
                ("exclusive owner, no need to notify other group members");
            exclusive = true;
            notify = false;
        } else {
            Global.debug ("exclusive owner, notifying other members");
            exclusive = true;
            notify = true;
        }

        if (exclusive)
            // arrange to be notified when somebody looks at the contents
            contents = new ExclusiveOwnership (contents);

        // set contents into ourselves
        super.setContents (contents, owner);

        // notify other group members
        if (notify) {
            Global.debug ("notifying other group members");
            Clipboard[] grp = getClipboards ();
            for (int i=0; i < grp.length; ++i) {
                Clipboard c = grp[i];
                c.setContents (contents, owner);
            }
        }

        flavors = newFlavors;
        String mimeType = Flavors.toMimeType (flavors);

        // notify system clipboard (unless it provided the contents)
        if (owner != SYSTEM_OWNER)
            acquireSystemClipboard (mimeType);

        Global.debug ("clipboard has become "
                                  + Flavors.getName (flavors));

        // tell UI what the type is
        Global.getUI ().setMimeType (mimeType);

        // display message in status bar
        Locator loc = (contents instanceof Locatable)
            ? ((Locatable)contents).getLocator ()
            : null;
        Global.getUI ().reportStatus (
            ((loc != null)
             ? loc.getHostName () + " copied "
             : "Copied ")
            + Flavors.getName (flavors)
            + " to clipboard.");

    }

    
    //
    // Wrapper around Transferable that notifies us when somebody looks
    // at its content, so we know when we no longer have exclusive ownership
    // over the content.
    //

    class ExclusiveOwnership implements Transferable {
        Transferable contents;
        public ExclusiveOwnership (Transferable contents) {
            this.contents = contents;
        }
        public Object getTransferData (DataFlavor flavor) throws UnsupportedFlavorException, IOException {
            try {
                return contents.getTransferData (flavor);
            } finally {
                if (exclusive) {
                    Global.debug ("giving up exclusive ownership");
                    exclusive = false;
                }
            }
        }
        public DataFlavor[] getTransferDataFlavors () {
            return contents.getTransferDataFlavors ();
        }
        public boolean isDataFlavorSupported (DataFlavor flavor) {
            return contents.isDataFlavorSupported (flavor);
        }
    };




    //
    // Overridden by subclass to acquire system clipboard on behalf
    // of synchronized clipboard.
    //
    protected abstract void acquireSystemClipboard (String mimeType);


    //
    // Specialized ClipboardOwners
    //
    
    // Passed to setContents() (usually by subclass) to indicate that
    // the system clipboard currently controls the synchronized clipboard.
    public static final ClipboardOwner SYSTEM_OWNER = new ClipboardOwner () {
        public void lostOwnership (Clipboard clipboard, 
                                   Transferable contents) {
        }
        public String toString () {
            return "system clipboard";
        }
    };


    
    // 
    // adding and removing members from clipboard group
    //

    public void add (Locator locator) {
        add (new ConnectingClipboard (locator));
    }

    public void add (Clipboard clipboard) {
        Locator locator =
            (clipboard instanceof Locatable)
            ? ((Locatable)clipboard).getLocator ()
            : null;

        if (myLocator.equals (locator))
            // don't add localhost
            return;

        boolean changed = false;

    search:
        synchronized (group) {
            for (int i = 0, n = group.size (); i < n; ++i) {
                Clipboard c = (Clipboard) group.elementAt (i);
 
                if (c.equals (clipboard)
                    || (locator != null 
                        && c instanceof Locatable
                        && locator.equals (((Locatable)c).getLocator ()))) {

                    // close old clipboard and replace it with new one
                    if (c instanceof Closable)
                        ((Closable)c).close ();
                    group.setElementAt (clipboard, i);
                    Global.getUI ().reportStatus (clipboard.getName () 
                                                  + " connected.");
                    changed = true;
                    break search;
                }
            }

            group.addElement (clipboard);
            changed = true;
            saveGroup ();
        }

        if (clipboard instanceof ConnectingClipboard)
            ((ConnectingClipboard)clipboard).connect ();

        if (changed)
            fireGroupChanged ();
    }

    public class ConnectingClipboard extends Clipboard implements Locatable, Runnable {
        Locator locator;
        boolean connecting = false;
        public ConnectingClipboard (Locator locator) {
            super (null);
            this.locator = locator;
        }
        public String getName () {
            return locator.getHumanName () + " (connecting)";
        }
        public void setContents (Transferable contents, ClipboardOwner owner) {
        }
        public Transferable getContents (Object requestor) {
            return null;
        }
        public Locator getLocator () {
            return locator;
        }
        public synchronized void connect () {
            if (!connecting) {
                connecting = true;
                new Thread (this).start ();
            }
        }
        public void run () {
            try {
                String msg = 
                    "Connecting to " 
                    + locator.getHumanName ()
                    + "...";
                int p = locator.getProtocol ();
                if (p == Locator.RMI || p == Locator.SECURE_RMI)
                    msg += " (check for authorization dialog there)";

                Global.getUI ().reportStatus (msg);
                protocol.join (locator);
            } catch (Exception e) {
                Global.report (e);
                Global.getUI ().reportStatus ("Connection to " 
                                       + locator.getHumanName ()
                                       + " failed: "
                                       + toReason (e)
                                       + ".");
                disconnect (this);
            }
        }
        String toReason (Throwable e) {
            String reason = e.getMessage ();
            if (reason == null || reason.length () == 0)
                reason = e.toString ();
            return reason;
        }
        public boolean equals (Object obj) {
            return (obj instanceof ConnectingClipboard
                    && ((ConnectingClipboard)obj).locator.equals (locator));
        }
    }

    public class DisconnectedClipboard extends Clipboard implements Locatable {
        Locator locator;
        public DisconnectedClipboard (Locator locator) {
            super (null);
            this.locator = locator;
        }
        public String getName () {
            return locator.getHumanName () + " (disconnected)";
        }
        public void setContents (Transferable contents, ClipboardOwner owner) {
        }
        public Transferable getContents (Object requestor) {
            return null;
        }
        public Locator getLocator () {
            return locator;
        }
        public boolean equals (Object obj) {
            return (obj instanceof DisconnectedClipboard
                    && ((DisconnectedClipboard)obj).locator.equals (locator));
        }
    }


    public boolean contains (Clipboard clipboard) {
        return group.contains (clipboard);
    }

    public void disconnect (Clipboard clipboard) {
	Global.debug ("disconnecting from " + clipboard);
        boolean changed = false;

        synchronized (group) {
            Global.debug ("clipboard group is currently " 
                                      + group);
            // Find the clipboard
            int i = group.indexOf (clipboard);
            if (i == -1) {
                Global.debug (clipboard + " not found");
                return;
            }

            // If clipboard has a locator (meaning it can be reopened in
            // the future), stick in a placeholder for it;
            // otherwise just delete it.
            if (clipboard instanceof Locatable) {
                Locator loc = ((Locatable)clipboard).getLocator ();
                Clipboard c = new DisconnectedClipboard (loc);
                group.setElementAt (c, i);
            } else 
                group.removeElementAt (i);
        }

        fireGroupChanged ();

        // close the clipboard
        if (clipboard instanceof Closable)
            ((Closable)clipboard).close ();
    }

    public void remove (Clipboard clipboard) {
        synchronized (group) {
            if (!group.removeElement (clipboard))
                return;
            saveGroup ();
        }
        if (clipboard instanceof Removable)
            ((Removable)clipboard).remove ();

        fireGroupChanged ();
    }
    
    void fireGroupChanged () {
        exclusive = false;
        Global.getUI ().groupChanged ();
    }

    //
    // enumerating clipboard group
    //

    public Enumeration clipboards () {
        return group.elements ();
    }

    public Clipboard[] getClipboards  () {
        synchronized (group) {
            Clipboard[] result = new Clipboard[group.size ()];
            for (int i = 0, n = group.size (); i < n; ++i)
                result[i] = (Clipboard) group.elementAt (i);
            return result;
        }
    }

    public Locator[] getLocators  () {
        synchronized (group) {
            Locator[] result = new Locator[group.size ()];
            for (int i = 0, n = group.size (); i < n; ++i) {
                Clipboard c = (Clipboard) group.elementAt (i);
                result[i] = (c instanceof Locatable)
                    ? ((Locatable)c).getLocator ()
                    : null;
            }                
            return result;
        }
    }

    //
    // making clipboard group persistent
    //

    public void saveGroup () {
        Global.getConfig ().removeAllKeysStartingWith ("Group");
        Locator[] locators = getLocators ();
        for (int i = 0; i < locators.length; ++i)
            if (locators[i] != null)
                Global.getConfig ().put ("Group" + i, locators[i].toString ());
        Global.getConfig ().save ();
    }

    public void restoreGroup () {
        Vector adds = new Vector ();
        for (Enumeration e = Global.getConfig ().propertyNames ();
             e.hasMoreElements (); ) {
            String name = (String)e.nextElement ();
            if (name.startsWith ("Group"))
                adds.addElement (Global.getConfig ().getProperty (name));
        }

        for (Enumeration e = adds.elements (); e.hasMoreElements (); )
            add (new Locator ((String) e.nextElement ()));
    }


}
