/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 2000 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

#include "WindowsClipboard.h"
#include <stdio.h>
#include <windows.h>
#include <shlobj.h>

/////////////////////////////
///////////////////////////// Internal state
/////////////////////////////

typedef struct _WindowsClipboardInternal {
    JavaVM *vm;             // Java virtual machine
    JNIEnv *globalEnv;      // Java environment (specific to ThreadBody's thread)
    jobject globalSelf;     // global reference to WindowsClipboard Java object
    HWND hwnd;              // ThreadBody's window (which acts as a clipboard viewer)
    HWND hwndNextCBChain;   // next window in clipboard viewer chain

    jmethodID GetSynchronizedClipboardData;
    jmethodID ClipboardChanged;

    jboolean debug;
} WindowsClipboardInternal;

/*
 * Get internal pointer from Java object.
 */
WindowsClipboardInternal* GetInternal (JNIEnv *env, jobject self)
{
    jclass cls;
    jfieldID field;

    // Get our class object
    cls = (*env)->GetObjectClass (env, self);
    //fprintf (stderr, "WindowsClipboard class=%p\n", cls);

    // Find the method ID for get()
    field = (*env)->GetFieldID (env, cls, "internal", "I");
    if (!field) {
        fprintf (stderr, "can't find internal\n");
        return NULL;
    }

    return (WindowsClipboardInternal*) (*env)->GetIntField (env, self, field);
}


/*
 * Set internal pointer in Java object.
 */
void SetInternal (JNIEnv *env, jobject self, WindowsClipboardInternal* internal)
{
    jclass cls;
    jfieldID field;

    // Get our class object
    cls = (*env)->GetObjectClass (env, self);
    //fprintf (stderr, "WindowsClipboard class=%p\n", cls);

    // Find the method ID for get()
    field = (*env)->GetFieldID (env, cls, "internal", "I");
    if (!field) {
        fprintf (stderr, "can't find internal\n");
        return;
    }

    (*env)->SetIntField (env, self, field, (jint)internal);
}


/////////////////////////////
///////////////////////////// Clipboard format conversions
/////////////////////////////

/*
 * Converting plain text clipboard format to a Java string
 */
jobject ConvertTextToJString (JNIEnv *env, HANDLE h)
{
    // Plain text
    char *s = (char*)GlobalLock (h);
    jobject result = (*env)->NewStringUTF (env, s);
    GlobalUnlock (h);
    return result;
}

/*
 * Converting HDROP clipboard format to an array of Java strings
 */
jobject ConvertDropFilesToJArray (JNIEnv *env, HANDLE h)
{
    HDROP hdrop = (HDROP)h;
    int nFiles;
    int i;
    char *buf;
    int maxBuf;
    jclass clazz;
    jobjectArray array;

    // Count the files
    nFiles = DragQueryFile (hdrop, -1, NULL, 0);

    // Allocate a Java array for nFiles strings
    clazz = (*env)->FindClass (env, "java/lang/String");
    if (!clazz) {
        fprintf (stderr, "Can't find java.lang.String\n");
        return NULL;
    }

    array = (*env)->NewObjectArray (env, nFiles, clazz, NULL);
    if (!clazz) {
        fprintf (stderr, "Can't allocate String array\n");
        return NULL;
    }

    // Allocate a filename buffer (can grow if needed)
    buf = malloc (maxBuf = 1024);

    // Loop through filenames, copying each one to the Java array
    for (i = 0; i < nFiles; ++i) {
        int maxFilename = DragQueryFile (hdrop, i, NULL, 0) + 1;
        if (maxFilename > maxBuf) {
            // Enlarge buffer
            if (buf != NULL)
                free(buf);
            buf = malloc (maxBuf = maxFilename);
        }

    	DragQueryFile (hdrop, i, buf, maxBuf);

        (*env)->SetObjectArrayElement (env, array, i, (*env)->NewStringUTF (env, buf));
    }

    // clean up
    free (buf);

    return array;
}

/*
 * Convert a java.lang.String to a Windows global handle.
 */
HANDLE ConvertJStringToText (JNIEnv *env, jstring str)
{
    const char *s;
    HANDLE h;
    char *t;

    // Get a pointer to the jstring's characters
    s = (*env)->GetStringUTFChars (env, str, 0);
    //fprintf (stderr, "get() returned '%s'\n", s);

    // Allocate a handle
	h = GlobalAlloc (GMEM_MOVEABLE | GMEM_DDESHARE, strlen(s)+1);

	// Copy jstring into the handle
	t = (char *)GlobalLock (h);
	strcpy (t, s);
	GlobalUnlock (h);

    // Release the jstring
    (*env)->ReleaseStringUTFChars (env, str, s);

	return h;
}

/*
 * Convert an array of java.lang.Strings to a DROPFILES structure.
 */
HANDLE ConvertJArrayToDropFiles (JNIEnv *env, jobjectArray array)
{
    HANDLE h;
    int i, n;
    int nBytes;
    DROPFILES *df;
    const char *s;
    char *t;

    // Measure size of DROPFILES structure needed
    n = (*env)->GetArrayLength (env, array);
    nBytes = sizeof (DROPFILES); // header
    for (i = 0; i < n; ++i) {
        jstring str = (*env)->GetObjectArrayElement (env, array, i);
        nBytes += (*env)->GetStringUTFLength (env, str) + 1;
    }
    nBytes += 1;    // extra null terminator marking end of file list

    // Allocate a handle
	h = GlobalAlloc (GMEM_MOVEABLE | GMEM_DDESHARE, nBytes);

    // Initialize the DROPFILES header
    df = (DROPFILES*) GlobalLock (h);
    df->pFiles = sizeof(DROPFILES);
    df->pt.x = 0;
    df->pt.y = 0;
    df->fNC = FALSE;
    df->fWide = FALSE;

    // Copy each filename into the DROPFILES buffer
    t = ((char*)df) + df->pFiles;
    for (i = 0; i < n; ++i) {
        jstring str = (*env)->GetObjectArrayElement (env, array, i);

        // Get a pointer to the jstring's characters
        s = (*env)->GetStringUTFChars (env, str, 0);

        // Append it to the DROPFILES buffer
        strcpy (t, s);
        t += strlen(t) + 1;

        // Release the jstring
        (*env)->ReleaseStringUTFChars (env, str, s);
    }

    // Add the final null terminator
    *t = '\0';

	GlobalUnlock (h);

	return h;
}

/*
 * Get a MIME type constant defined in Flavors.
 */

jstring GetMimeType (JNIEnv *env, const char *name) 
{
    jclass clsSource = (*env)->FindClass (env, "Flavors");
    jfieldID id = (*env)->GetStaticFieldID (env, clsSource, name,
                                            "Ljava/lang/String;");
    if (!id) {
        fprintf (stderr, "can't find Flavors.%s\n", name);
        return NULL;
    }

    return (*env)->GetStaticObjectField (env, clsSource, id);
}




/////////////////////////////
///////////////////////////// Window procedure
/////////////////////////////

// Window class name and title
#define WINDOW_CLASSNAME     TEXT("TransparentClipboard")
#define WINDOW_TITLE         TEXT("TransparentClipboard")


/*
 * Window procedure.
 */
BOOL CALLBACK WndProc (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
    BOOL lResult = TRUE;
    WindowsClipboardInternal *internal = (WindowsClipboardInternal*)GetWindowLong (hwnd, GWL_USERDATA);

    switch (msg) {
        case WM_CREATE:
            {
                CREATESTRUCT *cs = (CREATESTRUCT*)lp;
                internal = (WindowsClipboardInternal*)cs->lpCreateParams;

                if (internal->debug) fprintf (stderr, "WM_CREATE\n");

                // Store the internal pointer
                SetWindowLong (hwnd, GWL_USERDATA, (LONG)internal);

                // Add me to the clipboard viewer chain
                internal->hwndNextCBChain = SetClipboardViewer (hwnd);
                if (internal->debug) fprintf (stderr, "SetClipboardViewer() = %d\n", internal->hwndNextCBChain);
            }
            break;

        case WM_DESTROY:
            if (internal->debug) fprintf (stderr, "WM_DESTROY\n");

            // Unhook me from the clipboard viewer chain
            ChangeClipboardChain (hwnd, internal->hwndNextCBChain);

            // Tell the thread to quit
            PostQuitMessage (0);
            break;

        case WM_DRAWCLIPBOARD:
            if (internal->debug) fprintf (stderr, "WM_DRAWCLIPBOARD\n");
            fflush (stderr);

            // If somebody else acquired the clipboard, notify self using
            // windowsClipboardChanged ()
            if (GetClipboardOwner () != hwnd)
                (*internal->globalEnv)->CallVoidMethod (internal->globalEnv, internal->globalSelf, internal->ClipboardChanged);

            if (internal->hwndNextCBChain)
                SendMessage (internal->hwndNextCBChain, WM_DRAWCLIPBOARD, wp, lp);
            break;

        case WM_CHANGECBCHAIN:
            if (internal->debug) fprintf (stderr, "WM_CHANGECBCHAIN\n");

            if ((HWND)wp == internal->hwndNextCBChain)
                internal->hwndNextCBChain = (HWND)lp;
            else if (internal->hwndNextCBChain)
                SendMessage (internal->hwndNextCBChain, WM_CHANGECBCHAIN, wp, lp);
            break;

        case WM_RENDERALLFORMATS:
        case WM_RENDERFORMAT:
            if (internal->debug) {
		if (msg == WM_RENDERFORMAT)
			fprintf (stderr, "WM_RENDERFORMAT (%s)\n",
				(wp == CF_TEXT ? "CF_TEXT" : (wp == CF_HDROP ? "CF_HDROP" : "unknown")));
		else
			fprintf (stderr, "WM_RENDERALLFORMATS\n");
	    }
            {
                jobject obj;
                HANDLE h;

                // Get the network clipboard contents
            	obj = (*internal->globalEnv)->CallObjectMethod (internal->globalEnv, internal->globalSelf, internal->GetSynchronizedClipboardData);
            	if (!obj) {
            	    fprintf (stderr, "get() returned null\n");
            	    break;
            	}

                if ((*internal->globalEnv)->IsInstanceOf (internal->globalEnv, obj,
                        (*internal->globalEnv)->FindClass (internal->globalEnv, "java/lang/String"))) {
                    // Convert string to a handle
                	h = ConvertJStringToText (internal->globalEnv, obj);
                	if (!h) {
                	    fprintf (stderr, "couldn't allocate handle for text");
                	    break;
                	}

                    // Put the text on the clipboard, as requested
                	SetClipboardData (CF_TEXT, h);
                }
                else {
                    // Assume it's an array of Strings (corresponding to filenames)
                    h = ConvertJArrayToDropFiles (internal->globalEnv, obj);
                	if (!h) {
                	    fprintf (stderr, "couldn't allocate handle for dropped files");
                	    break;
                	}

                    // Put the text on the clipboard, as requested
                	SetClipboardData (CF_HDROP, h);
                }
            }
        	break;

        default:
            lResult = DefWindowProc(hwnd, msg, wp, lp);
            break;
    }

    return lResult;
}

/*
 * Message loop thread.
 */
DWORD ThreadBody (WindowsClipboardInternal *internal)
{
    HINSTANCE hInstance = GetModuleHandle (NULL);
    WNDCLASS wc;
    MSG msg;
    int result;

    // Attach this thread to the VM
    result = (*internal->vm)->AttachCurrentThread (internal->vm, (void*)&internal->globalEnv, NULL);
    if (internal->debug) fprintf (stderr, "AttachCurrentThread returned result=%d, globalEnv=%p\n", result, internal->globalEnv);

    if (internal->debug) fprintf (stderr, "WindowsClipboard_createInternal\n");

    wc.style = 0;
    wc.lpfnWndProc = (WNDPROC)WndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hIcon = NULL;
    wc.hInstance = hInstance;
    wc.hCursor = NULL;
    wc.hbrBackground = (HBRUSH) GetStockObject( WHITE_BRUSH );
    wc.lpszMenuName = NULL;
    wc.lpszClassName = WINDOW_CLASSNAME;

    if (!RegisterClass(&wc)) {
        fprintf (stderr, "WindowsClipboard: can't register window class\n");
        return 0;
    }

    internal->hwnd = CreateWindow (WINDOW_CLASSNAME,
                                   WINDOW_TITLE,
                                   0,
                                   CW_USEDEFAULT,
                                   CW_USEDEFAULT,
                                   CW_USEDEFAULT,
                                   CW_USEDEFAULT,
                                   NULL, NULL, hInstance, (LPVOID)internal );
    if (!internal->hwnd) {
        fprintf (stderr, "WindowsClipboard: can't create window\n");
        return 0;
    }

    // Message pump
    while (GetMessage (&msg, NULL, 0, 0)) {
        TranslateMessage (&msg);
        DispatchMessage (&msg);
    }

    // Free the global WindowsClipboard ref
    (*internal->globalEnv)->DeleteGlobalRef (internal->globalEnv, internal->globalSelf);
    internal->globalSelf = (jobject)0;

    // Detach this thread from the VM, making globalEnv invalid
    (*internal->vm)->DetachCurrentThread (internal->vm);
    internal->globalEnv = NULL;

    // free the internal structure
    free (internal);

    return msg.wParam;
}


/////////////////////////////
///////////////////////////// WindowsClipboard native method implementations
/////////////////////////////


/*
 * Class:     WindowsClipboard
 * Method:    openWindowsClipboard
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WindowsClipboard_openWindowsClipboard
  (JNIEnv * env, jobject self, jboolean dbg)
{
    WindowsClipboardInternal *internal;
    jclass cls;
    DWORD idThread;

    if (dbg) {
        fprintf (stderr, "WindowsClipboard_openWindowsClipboard\n");
        fflush (stderr);
    }

    internal = malloc (sizeof(WindowsClipboardInternal));
    if (!internal) {
        fprintf (stderr, "Can't allocate internal structure\n");
        return;
    }
    SetInternal (env, self, internal);
    internal->debug = dbg;

    (*env)->GetJavaVM (env, &internal->vm);

    internal->globalSelf = (*env)->NewGlobalRef (env, self);
    if (!internal->globalSelf)
        fprintf (stderr, "Can't create globalSelf\n");
    if (internal->debug) fprintf (stderr, "self=%p, globalSelf=%p\n", self, internal->globalSelf);

    // Find the method IDs we'll need
    cls = (*env)->GetObjectClass (env, self);
    if (internal->debug) fprintf (stderr, "WindowsClipboard class=%p\n", cls);

    internal->ClipboardChanged = (*env)->GetMethodID (env, cls, "windowsClipboardChanged", "()V");
    if (!internal->ClipboardChanged)
        fprintf (stderr, "can't find windowsClipboardChanged()\n");

    internal->GetSynchronizedClipboardData = (*env)->GetMethodID (env, cls, "getSynchronizedClipboardData", "()Ljava/lang/Object;");
    if (!internal->GetSynchronizedClipboardData)
        fprintf (stderr, "can't find getSynchronizedClipboardData()\n");

    if (!CreateThread (NULL, 0, (LPTHREAD_START_ROUTINE)ThreadBody, (LPVOID)internal, 0, &idThread))
        fprintf (stderr, "Can't create thread\n");

    fflush (stderr);
}

/*
 * Class:     WindowsClipboard
 * Method:    closeWindowsClipboard
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WindowsClipboard_closeWindowsClipboard
  (JNIEnv * env, jobject self)
{
    WindowsClipboardInternal *internal = GetInternal (env, self);

    if (internal->debug) fprintf (stderr, "WindowsClipboard_closeWindowsClipboard\n");
    DestroyWindow (internal->hwnd);
}

/*
 * Class:     WindowsClipboard
 * Method:    getWindowsClipboardData
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jobject JNICALL Java_WindowsClipboard_getWindowsClipboardData
  (JNIEnv * env, jobject self)
{
    WindowsClipboardInternal *internal = GetInternal (env, self);
    jobject result = NULL;

    if (internal->debug) fprintf (stderr, "WindowsClipboard_getWindowsClipboardData\n");
    if (OpenClipboard (internal->hwnd)) {
        HANDLE h;

        if ((h = (HANDLE)GetClipboardData (CF_HDROP)) != 0)
            result = ConvertDropFilesToJArray (env, h);
        else if ((h = (HANDLE)GetClipboardData (CF_TEXT)) != 0)
            result = ConvertTextToJString (env, h);
        else
            fprintf (stderr, "Clipboard not text\n");

    	CloseClipboard ();
    }
    else
        fprintf (stderr, "Clipboard busy\n");

    return result;
}

/*
 * Class:     WindowsClipboard
 * Method:    getWindowsMimeType
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_WindowsClipboard_getWindowsMimeType
  (JNIEnv * env, jobject self)
{
    WindowsClipboardInternal *internal = GetInternal (env, self);
    const char *typename = NULL;

    if (OpenClipboard (internal->hwnd)) {
        UINT fmt = 0;

        // Enumerate all formats on clipboard looking for one we recognize
        while (!typename
                && (fmt = EnumClipboardFormats (fmt)) != 0) {
            if (fmt == CF_HDROP) {
                typename = "FILES_TYPE";
                break;
            }
            else if (fmt == CF_TEXT) {
                typename = "STRING_TYPE"; // Text
                break;
            }
        }

        CloseClipboard ();
    }
    else
        fprintf (stderr, "Clipboard busy\n");

    if (!typename)
        typename = "UNKNOWN_TYPE";

    return GetMimeType (env, typename);
}

/*
 * Class:     WindowsClipboard
 * Method:    acquireSystemClipboard
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WindowsClipboard_acquireSystemClipboard
  (JNIEnv * env, jobject self, jstring jContentType) {
    WindowsClipboardInternal *internal = GetInternal (env, self);
    UINT fmt;
    const char *contentType = (*env)->GetStringUTFChars (env, jContentType, 0);

    if (!strcmp (contentType, "application/x-filenames"))
        fmt = CF_HDROP;
    else if (!strcmp (contentType, "text/plain"))
        fmt = CF_TEXT;
    else
        fmt = 0;

    (*env)->ReleaseStringUTFChars (env, jContentType, contentType);

    if (!fmt)
        return;

    if (OpenClipboard (internal->hwnd)) {
    	EmptyClipboard ();
    	SetClipboardData (fmt, NULL);
    	CloseClipboard ();
    }
    else
        fprintf (stderr, "Clipboard busy\n");
}
