/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.awt.datatransfer.*;
import java.io.*;
import rcm.awt.PopupDialog;
import rcm.io.TempFile;
import rcm.util.Str;

public class WindowsClipboard extends SynchronizedClipboard {
    int internal;  // pointer to internal data
    FileList tempFiles;   
            // temporary files created by last file-paste,
            // or null if last paste was not files

    public WindowsClipboard () {
        System.loadLibrary ("WindowsClipboard");        
        Global.debug ("opening Windows clipboard");
        openWindowsClipboard (Global.getDebug ().isEnabled ());
    }
    
    public void close () {
        super.close ();
        deleteTempFiles ();
        closeWindowsClipboard ();
    }

    void deleteTempFiles () {
        if (tempFiles != null) {
            try {
                tempFiles.delete ();
            } catch (Exception e) {
            }
            tempFiles = null;
        }
    }

    //
    // Transferable that fetches data from system clipboard
    //

    Transferable systemClipboardContents = new Transferable () {
        public Object getTransferData (DataFlavor flavor) throws UnsupportedFlavorException, IOException {
            enter ();
            try {
                Object obj = getWindowsClipboardData ();
                if (flavor.equals (Flavors.filesFlavor)) {
                    if (obj instanceof String[])
                    // obj is a list of filenames -- convert them to a FileList
                    return new FileList ((String[]) obj);
                } else if (flavor.equals (DataFlavor.stringFlavor)) {
                    if (obj instanceof String)
                      // Convert CR/LF to LF
                      return Str.replace ((String)obj, "\r\n", "\n");
                } else if (flavor.equals (DataFlavor.plainTextFlavor)) {
                    if (obj instanceof String)
                      // Convert CR/LF to LF
                      return new StringReader (Str.replace ((String)obj, "\r\n", "\n"));
                }

                throw new UnsupportedFlavorException (flavor);
            } finally {
                leave ();
            }
        }

        public DataFlavor[] getTransferDataFlavors () {
            enter ();
            try {
                return Flavors.toDataFlavors (getWindowsMimeType ());
            } finally {
                leave ();
            }
        }

        public boolean isDataFlavorSupported (DataFlavor flavor) {
            enter ();
            try {
                return Flavors.isDataFlavorSupported (flavor,
                                                      getTransferDataFlavors ());
            } finally {
                leave ();
            }
        }

        // Check for infinite recursion.  This isn't supposed to
        // happen, but programming error could cause system clipboard
        // and synchronized clipboard to reference each other.
        boolean recursed = false;
        void enter () {
            if (recursed)
                throw new RuntimeException ("infinite recursion");
            recursed = true;
        }
        void leave () {
            recursed = false;
        }
    };

    //
    // Native interface (implemented in WindowsClipboard.dll)
    //

    // Initialize/cleanup DLL
    native void openWindowsClipboard (boolean debug);
    native void closeWindowsClipboard ();

    // Fetch data from Windows clipboard.  Returns String for text, 
    // String[] for files, or null for anything else.
    native Object getWindowsClipboardData ();

    // Fetch type of data on Windows clipboard.  Returns STRING_TYPE,
    // FILES_TYPE, or UNKNOWN_TYPE.
    native String getWindowsMimeType ();

    // Acquire Windows clipboard on behalf of synchronized clipboard.
    // mimeType is STRING_TYPE, FILES_TYPE, or UNKNOWN_TYPE.
    protected native void acquireSystemClipboard (String mimeType);
    
    //
    // Up-call interface (called by WindowsClipboard.dll)
    //

    // Called by DLL when Windows clipboard changes.
    void windowsClipboardChanged () {
        Global.debug ("another app put data on Windows clipboard");
        setContents (systemClipboardContents, SYSTEM_OWNER);
    }
    
    // Called by DLL when Windows clipboard demands data from
    // synchronized clipboard.
    Object getSynchronizedClipboardData () {
        // Delete temporary files from previous paste, if any
        deleteTempFiles ();

        // Fetch the clipboard contents
        Transferable t = getContents (null);

        // First try to interpret clipboard as a file list
        try {
            FileList fl = (FileList) t.getTransferData (Flavors.filesFlavor);

            if (fl instanceof RemoteFileList) {
                // Copy remote files to temporary directory on local machine
                try {
                    fl = ((RemoteFileList)fl).makeLocal ();
                    tempFiles = fl;
                } catch (IOException e) {
                    PopupDialog.warn (null, "Error", e.toString ());
                    Global.getUI ().reportStatus ("Paste failed.");
                    return null;
                }
            }

            // Convert FileList to a simple array of Strings; native
            // code will further convert it to Windows file-list clipboard 
            // format.
            // Remove trailing slashes (Windows clipboard doesn't like 'em)
            String[] pastedFiles = fl.getPathnames ();
            for (int i = 0; i < pastedFiles.length; ++i) {
                String filename = pastedFiles[i];
                filename = Str.replace (filename, "/", "\\");
                if (filename.endsWith ("\\"))
                    filename = filename.substring (0, filename.length()-1);
                pastedFiles[i] = filename;
            }

            Global.getUI ().reportStatus ("Pasted files from clipboard.");
            return pastedFiles;
        } catch (UnsupportedFlavorException e) {
            Global.report (e);
        } catch (IOException e) {
            Global.report (e);
        }

        // Next try getting a string
        try {
            String str = (String) t.getTransferData (DataFlavor.stringFlavor);
            // Convert LF to CR/LF
            str = Str.replace (str, "\n", "\r\n");
            Global.getUI ().reportStatus ("Pasted text from clipboard.");
            return str;
        } catch (UnsupportedFlavorException e) {
            Global.report (e);
        } catch (IOException e) {
            Global.report (e);
        }

        Global.getUI ().reportStatus ("Paste failed: nothing recognizable on clipboard.");
        return null;
    }
}
