/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 2000 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#include <stdio.h>
#include "XClipboard.h"

jboolean debug;

jmethodID GetSynchronizedClipboardData;
jmethodID ClipboardChanged;

Atom clipboardAtom, primaryAtom;
Atom stringAtom, textAtom, compoundTextAtom;
Atom targetsAtom, multipleAtom, timestampAtom;
Atom remoteClipAtom;
Atom managerAtom;

Atom lastSelection;

jboolean isText; // if contents of remote clipboard is text

Display *globalDisplay;
Window globalWin;


/*
 * JNI interface
 */


/*
 * Class:     XClipboard
 * Method:    openXClipboard
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_XClipboard_openXClipboard
    (JNIEnv *env, jobject self, jboolean dbg) {
    Display *display;
    int screen_num;
    Window win;
    jclass cls;

    debug = dbg;

    if (debug)
        fprintf (stderr, "opening X clipboard\n");

    /* Find the Java method IDs we'll need */
    cls = (*env)->GetObjectClass (env, self);

    ClipboardChanged = (*env)->GetMethodID (env, cls, 
                                            "XClipboardChanged", 
                                            "()V");
    if (!ClipboardChanged)
        fprintf (stderr, "can't find windowsClipboardChanged()\n");

    GetSynchronizedClipboardData = 
        (*env)->GetMethodID (env, cls, 
                             "getSynchronizedClipboardData", 
                             "()Ljava/lang/String;");
    if (!GetSynchronizedClipboardData)
        fprintf (stderr, "can't find getSynchronizedClipboardData()\n");

    /* connect to X server */
    if ((display = XOpenDisplay (NULL)) == NULL) {
        fprintf (stderr, "can't connect to X server %s\n",
                 XDisplayName (NULL));
        return;
    }

    /* get atoms */
    clipboardAtom = XInternAtom (display, "CLIPBOARD", False);
    primaryAtom = XInternAtom (display, "PRIMARY", False);
    stringAtom = XInternAtom (display, "STRING", False);
    textAtom = XInternAtom (display, "TEXT", False);
    compoundTextAtom = XInternAtom (display, "COMPOUND_TEXT", False);
    targetsAtom = XInternAtom (display, "TARGETS", False);
    multipleAtom = XInternAtom (display, "MULTIPLE", False);
    timestampAtom = XInternAtom (display, "TIMESTAMP", False);
    remoteClipAtom = XInternAtom (display, "_REMOTE_CLIP", False);

    lastSelection = primaryAtom;

    /* create hidden window */
    screen_num = DefaultScreen (display);
    win = XCreateSimpleWindow 
        (display, RootWindow (display, screen_num),
         0, 0, 1, 1, 1, BlackPixel (display, screen_num),
         WhitePixel (display, screen_num));
    
    /* select only unblockable events (like SelectionNotify) */
    XSelectInput (display, win, 0);

    /* ensure that we've contacted the server */
    XGetSelectionOwner (display, lastSelection);

    /* inhibit xclipboard, if it's running */
    managerAtom = XInternAtom(display, "CLIPBOARD_MANAGER", False);
    XSetSelectionOwner (display, managerAtom, win, CurrentTime);

    /* store the display and window */
    globalDisplay = display;
    globalWin = win;
}

/*
// Clipboard managers (like xclipboard or klipper)
// poll the clipboard in order to maintain a cut-and-paste history.
// If rclip is running on two machines with clipboard managers, this polling
// would force the machines to trade clipboard ownership back and forth.
// We handle this by simply rejecting requests from a clipboard manager.  
// As a result, remote selections won't appear in the history, 
// but local selections will still work as usual. 
*/
jboolean isClipboardManager (Display *display, Window requestor) {
    char *windowName = NULL;
    jboolean result;

    XFetchName (display, requestor, &windowName);
    result = (windowName 
              && (!strcmp (windowName, "klipper")
                  || !strcmp (windowName, "xclipboard")));
    XFree (windowName);

    return result;
}

Atom sendClipboardData (JNIEnv *env, jobject self,
                        Display *display, Window requestor, 
                        Atom target, Atom property)
{
    jstring jstr;
    const char *data;
    jboolean sendNothing = !isText || isClipboardManager (display, requestor);

/*     if (debug) { */
/*         fprintf (stderr, */
/*                  "selection requested by %d, format %s, property %s\n", */
/*                  requestor, */
/*                  XGetAtomName (display, target),  */
/*                  XGetAtomName (display, property)); */
/*         fflush (stderr); */
/*     } */

    if (property == None)
        property = target;

    /* handle TARGETS request in all cases */
    if (target == targetsAtom) {
        Atom targets[6];
        int nTargets;

        targets[0] = targetsAtom;
        targets[1] = multipleAtom;
        targets[2] = timestampAtom;
        if (sendNothing)
            nTargets = 3;
        else {
            targets[3] = stringAtom;
            targets[4] = textAtom;
            targets[5] = compoundTextAtom;
            nTargets = 6;
        }
        XChangeProperty (display, requestor, property,
                         XA_ATOM, 32, PropModeReplace,
                         (unsigned char *) targets, nTargets);
        return property;
    } 

    /* if sendNothing, refuse all other requests */
    if (sendNothing)
        return None;

    /* refuse all non-text requests */
    if (target != stringAtom
        && target != textAtom
        && target != compoundTextAtom) {
        fprintf (stderr, "unsupported format\n");
        return None;
    }
    
    /* Otherwise we want to send back the text of the clipboard.
       Call up to Java to get clipboard contents. */
    jstr = (*env)->CallObjectMethod 
        (env, self, GetSynchronizedClipboardData);
    
    if (!jstr) {
        fprintf (stderr, "getSynchronizedClipboardData() returned null\n");
        return None;
    }

    /* convert Java string to C string */
    data = (*env)->GetStringUTFChars (env, jstr, 0);

    /* store string in property */
    if (debug)
        fprintf (stderr, "storing text in property\n");
    XChangeProperty (display, requestor, property,
                     target == compoundTextAtom 
                     ? compoundTextAtom 
                     : stringAtom, 
                     8, PropModeReplace,
                     data, strlen(data));

    /* free C string */
    (*env)->ReleaseStringUTFChars (env, jstr, data);

    return property;
}


/*
 * Class:     XClipboard
 * Method:    runXClipboard
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_XClipboard_runXClipboard
    (JNIEnv *env, jobject self) {
    Display *display = globalDisplay;
    Window win = globalWin;
    XEvent event;
    XEvent notify;

    if (!display) {
        fprintf (stderr, "can't serve X selection without X connection\n");
        return;
    }

    /* event loop */
    if (debug)
        fprintf (stderr, "entering event loop\n");
    while (1) {
        XNextEvent (display, &event);
        switch (event.type) {
        case SelectionRequest: 
/*             if (debug) { */
/*                 fprintf (stderr, "SelectionRequest\n"); */
/*                 fflush (stderr); */
/*             } */
            /* send a SelectionNotify event to the requestor */
            notify.type = SelectionNotify;
            notify.xselection.serial = 0;
            notify.xselection.send_event = True;
            notify.xselection.display = 0;
            notify.xselection.requestor = event.xselectionrequest.requestor;
            notify.xselection.selection = event.xselectionrequest.selection;
            notify.xselection.target = event.xselectionrequest.target;
            notify.xselection.property = 
                sendClipboardData (env, self, display,
                                   event.xselectionrequest.requestor, 
                                   event.xselectionrequest.target, 
                                   event.xselectionrequest.property);
            notify.xselection.time = event.xselectionrequest.time;
/*             if (debug) */
/*                 fprintf (stderr, "sending SelectionNotify\n"); */
            XSendEvent (display, event.xselectionrequest.requestor, 
                        True, 0, &notify);
            break;

        case SelectionClear:
            if (debug) {
                fprintf (stderr, "lost selection\n");
                fflush (stderr);
            }
            lastSelection = event.xselectionclear.selection;
            (*env)->CallVoidMethod (env, self, ClipboardChanged);
            if (debug)
                fprintf (stderr, "returned from ClipboardChanged\n");
            break;

        case ClientMessage:
            if (debug) {
                fprintf (stderr, "grabbing X clipboard ownership\n");
                fflush (stderr);
            }
            XSetSelectionOwner (display, clipboardAtom, win, CurrentTime);
            XSetSelectionOwner (display, primaryAtom, win, CurrentTime);
            break;

        default:
            fprintf (stderr, "unexpected event: %d\n", event.type);
            fflush (stderr);
            break;
        }
    }
}


/*
 * Class:     XClipboard
 * Method:    acquireSystemClipboard
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_XClipboard_acquireSystemClipboard
    (JNIEnv *env, jobject self, jstring jContentType) 
{
    Display *display;
    XEvent notify;

    const char *contentType = (*env)->GetStringUTFChars (env, jContentType, 0);
    isText = (!strcmp (contentType, "text/plain"));
    (*env)->ReleaseStringUTFChars (env, jContentType, contentType);

    /* make a fresh connection to X server */
    if ((display = XOpenDisplay (NULL)) == NULL) {
        fprintf (stderr, "can't connect to X server %s\n",
                 XDisplayName (NULL));
        return;
    }

    /* tell hidden RemoteClip window to grab the selection */
    if (debug)
        fprintf (stderr, "sending ClientMessage\n");
    notify.type = ClientMessage;
    notify.xclient.serial = 0;
    notify.xclient.send_event = True;
    notify.xclient.display = 0;
    notify.xclient.window = globalWin;
    notify.xclient.message_type = None;
    notify.xclient.format = 8;
    XSendEvent (display, globalWin, True, 0, &notify);

    /* close up the temporary connection */
    XCloseDisplay (display);
}


/*
 * Class:     XClipboard
 * Method:    getXClipboardData
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_XClipboard_getXClipboardData
    (JNIEnv *env, jobject self) 
{
    Display *display;
    int screen_num;
    Window win;
    XEvent event;
    unsigned char *value;
    jstring jstr;

    if (debug)
        fprintf (stderr, "getting X clipboard data\n");

    /* make a fresh connection to X server */
    if ((display = XOpenDisplay (NULL)) == NULL) {
        fprintf (stderr, "can't connect to X server %s\n",
                 XDisplayName (NULL));
        return NULL;
    }

    screen_num = DefaultScreen (display);

    /* create hidden window */
    if (debug)
        fprintf (stderr, "creating window\n");
    win = XCreateSimpleWindow (display, RootWindow (display, screen_num),
                               0, 0, 1, 1, 1, BlackPixel (display, screen_num),
                               WhitePixel (display, screen_num));

    /* select only unblockable events (like SelectionNotify) */
    XSelectInput (display, win, 0);

    /* request selection */
    if (debug)
        fprintf (stderr, "requesting selection\n");
    XConvertSelection (display, lastSelection, 
                       stringAtom, remoteClipAtom,
                       win, CurrentTime);

    /* event loop */
    if (debug)
        fprintf (stderr, "getting an event\n");
    value = NULL;
    XNextEvent (display, &event);
    switch (event.type) {
    case SelectionNotify:
        if (event.xselection.property == None) {
            if (debug)
                fprintf (stderr, "no clipboard contents\n");
            break;
        } else {
            Atom actualType;
            int actualFormat;
            unsigned long nItemsReturned, bytesAfterReturn;
            
            // fetch the length of the clipboard
            if (XGetWindowProperty (display, win, event.xselection.property,
                                    0, 0, False, stringAtom,
                                    &actualType, &actualFormat,
                                    &nItemsReturned, &bytesAfterReturn,
                                    &value) != Success
                || actualType != stringAtom) {
                fprintf (stderr, "window property not found\n");
                break;
            }
            
            if (XGetWindowProperty (display, win, event.xselection.property,
                                    0, bytesAfterReturn, True, stringAtom,
                                    &actualType, &actualFormat,
                                    &nItemsReturned, &bytesAfterReturn,
                                    &value) != Success) {
                fprintf (stderr, "couldn't fetch window property\n");
                break;
            }
        }
        break;
    default:
        fprintf (stderr, "unexpected event: %d\n", event.type);
        break;
    }

    /* convert clipboard contents to a Java string */
    if (value) {
        if (debug)
            fprintf (stderr, "converting to Java string\n");
        jstr = (*env)->NewStringUTF (env, value);
        XFree (value);
    } else
        jstr = NULL;

    /* close up the temporary connection */
    XCloseDisplay (display);

    if (debug)
        fprintf (stderr, "returning Java string\n");
    return jstr;
}
