/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 1999-2001 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

import java.awt.datatransfer.*;
import java.io.*;
import rcm.util.Str;

public class XClipboard extends SynchronizedClipboard {
    public XClipboard (String libraryName) {
        System.loadLibrary (libraryName);

        openXClipboard (Global.getDebug ().isEnabled ());

        new Thread () {
            public void run () {
                runXClipboard ();
            }
        }.start ();

        seizeXClipboard ();
    }

    //
    // Transferable that fetches data from system clipboard
    //

    Transferable systemClipboardContents = new Transferable () {
        public Object getTransferData (DataFlavor flavor) throws UnsupportedFlavorException, IOException {
            enter ();
            try {
                Object obj = seizeXClipboard ().getTransferData (flavor);
                Global.debug ("systemClipboard.getTransferData returning " + obj);
                return obj;
            } finally {
                leave ();
            }
        }

        public DataFlavor[] getTransferDataFlavors () {
            return Flavors.STRING_FLAVORS;
        }

        public boolean isDataFlavorSupported (DataFlavor flavor) {
            return Flavors.isDataFlavorSupported (flavor,
                                                  Flavors.STRING_FLAVORS);
        }

        // Check for infinite recursion.  This isn't supposed to
        // happen, but programming error could cause system clipboard
        // and synchronized clipboard to reference each other.
        boolean recursed = false;
        void enter () {
            if (recursed)
                throw new RuntimeException ("infinite recursion");
            recursed = true;
        }
        void leave () {
            recursed = false;
        }
    };

    // When synchronized clipboard is set to an external
    // contents (like a remote clipboard), we assert ownership
    // of the system clipboard and put syncClipboardContents on it. 
    //
    // Native interface
    //

    // Initialize library
    native void openXClipboard (boolean debug);

    // Run event loop
    native void runXClipboard ();

    // Acquire X clipboard on behalf of synchronized clipboard.
    protected native void acquireSystemClipboard (String mimeType);
    
    // Acquire X clipboard on behalf of synchronized clipboard.
    native String getXClipboardData ();
    
    // Seize ownership of X clipboard, also obtaining its current contents
    protected Transferable seizeXClipboard () {
        Global.debug ("seizing ownership of X clipboard");
        String str = getXClipboardData ();
        if (str == null)
            str = "";
        Global.debug ("X clipboard was '" + str + "'");

        Transferable c = new StringSelection (str);
        setContents (c, null);
        Global.debug ("seizeXClipboard returning " + c);
        return c;
    }
    
    //
    // Up-call interface (called by XClipboard native code)
    //

    // Called by native code when X clipboard changes.
    void XClipboardChanged () {
        Global.debug ("another app put data on X clipboard");
        setContents (systemClipboardContents, SYSTEM_OWNER);
    }
    
    // Called by native code when X clipboard demands data from
    // synchronized clipboard.
    String getSynchronizedClipboardData () {
        // Fetch the clipboard contents
        Transferable t = getContents (null);

        // Try getting a string
        try {
            String str = (String) t.getTransferData (DataFlavor.stringFlavor);
            Global.getUI ().reportStatus ("Pasted text from clipboard.");
            return str;
        } catch (UnsupportedFlavorException e) {
            Global.report (e);
        } catch (IOException e) {
            Global.report (e);
        }

        Global.getUI ().reportStatus ("Paste failed: nothing recognizable on clipboard.");
        return null;
    }
}
