/*
 * LAPIS utility classes
 * Copyright (C) 1998-2002 Carnegie Mellon University. All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU Library
 * General Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */
package rcm.progress;

import java.awt.*;
import java.awt.event.*;
import rcm.util.Win;

public class AWTProgressBar extends Panel implements ProgressListener {

    Label label;
    Thermometer therm;
    boolean cancelled = false;
    Dialog dialog = null;

    public AWTProgressBar () {
        setLayout (new BorderLayout ());
        add (label = new Label (), BorderLayout.NORTH);
        add (therm = new Thermometer (), BorderLayout.SOUTH);
    }

    public Dialog makeDialog (Component parent, String title, boolean modal) {
        if (dialog != null)
            return dialog;

        dialog = new Dialog (Win.findFrameOrMakeFrame (parent),
                                 title, modal);

        dialog.setLayout (new BorderLayout ());
        dialog.add (this);

        Panel bottom = new Panel ();
        Button cancelButton = new Button ("Cancel");
        cancelButton.addActionListener (new ActionListener () {
            public void actionPerformed (ActionEvent event) {
                cancel ();
            }
        });
        bottom.add (cancelButton);
        dialog.add (bottom, BorderLayout.SOUTH);

        dialog.pack ();
        Win.center (dialog, parent);

        return dialog;
    }

    public void cancel () {
        cancelled = true;
        if (dialog != null)
            dialog.dispose ();
    }

    public Label getLabel () {
        return label;
    }

    public Thermometer getThermometer () {
        return therm;
    }

    public void startProgress (ProgressEvent event) {
        handle (event);
    }

    public void makeProgress (ProgressEvent event) {
        handle (event);
    }

    public void doneProgress (ProgressEvent event) {
        handle (event);
    }

    private void handle (ProgressEvent event) {
        if (cancelled)
            throw new CancelException ();
        label.setText (event.getLabel ());
        therm.setPercent (event.getPercentComplete ());
        if (dialog != null)
            dialog.update (dialog.getGraphics ());
        else
            update (getGraphics ());
    }

    public static class Thermometer extends Canvas {
        Color color = Color.blue;
        double percent = 0.0;
        Dimension preferredSize = new Dimension (100, 14);
        final static int margin = 0;

        public Thermometer () {
        }

        public void setColor (Color color) {
            this.color = color;
            repaint ();
        }
        
        public Color getColor () {
            return color;
        }
        
        public void setPercent (double percent) {
            this.percent = percent;
            repaint ();
        }

        public double getPercent () {
            return percent;
        }

        public Dimension getPreferredSize () {
            return preferredSize;
        }
        
        public void setPreferredSize (Dimension d) {
            preferredSize = new Dimension (d);
            Thermometer.this.invalidate ();
        }

        public void paint (Graphics g) {
            Dimension size = getSize ();

            // Draw black frame around thermometer
            int frameX = margin;
            int frameY = margin;
            int frameW = size.width - 1;
            int frameH = size.height - 1;
            g.setColor (Color.black);
            g.drawRect (frameX, frameY, frameW, frameH);

            // Fill in color
            int fillX = frameX + 1;
            int fillY = frameY + 1;
            int fillW = (int) ((frameW - 1) * percent);
            int fillH = frameH - 1;
            g.setColor (color);
            g.fillRect (fillX, fillY, fillW, fillH);
        }
    }


    public static void main (String[] args) throws Exception {
        AWTProgressBar pb = new AWTProgressBar ();
        Dialog dlg = pb.makeDialog (null, "Progress", false);
        dlg.show ();

        BasicProgressGenerator gen = new BasicProgressGenerator (dlg);
        gen.addProgressListener (pb);
        gen.fireStartProgress ("Counting to infinity");

        try {
            for (int i = 0; dlg != null && i < 20; ++i) {
                Thread.sleep (1000);
                gen.fireProgress ();
            }
            gen.fireDoneProgress ();
        } catch (CancelException e) {
            System.err.println (e);
        } finally {
            dlg.dispose ();
        } 
        
        System.exit (0);
    }

}
