#include <windows.h>
#include "TrayIcon.h"

#define ERROR_PRINT0(arg0)           (fprintf (stderr, arg0), fflush (stderr))
#define ERROR_PRINT1(arg0,arg1)      (fprintf (stderr, arg0, arg1), fflush (stderr))
#define ERROR_PRINT2(arg0,arg1,arg2) (fprintf (stderr, arg0, arg1, arg2), fflush (stderr))

#ifdef DEBUG

#define DEBUG_PRINT0(arg0)           (fprintf (stderr, arg0), fflush (stderr))
#define DEBUG_PRINT1(arg0,arg1)      (fprintf (stderr, arg0, arg1), fflush (stderr))
#define DEBUG_PRINT2(arg0,arg1,arg2) (fprintf (stderr, arg0, arg1, arg2), fflush (stderr))

#else

#define DEBUG_PRINT0(arg0)
#define DEBUG_PRINT1(arg0,arg1)
#define DEBUG_PRINT2(arg0,arg1,arg2)

#endif


typedef struct _TrayIconInternal {
    JavaVM *vm;             // Java virtual machine
    JNIEnv *env;            // Java environment for our thread
    jobject this;           // global reference to TrayIcon Java object
    HWND hwnd;              // our window
    NOTIFYICONDATA nid;     // tray icon data

    // method handles
    jmethodID internalMouseClickedID;

    // field handles
    jfieldID visibleID;
    jfieldID tooltipID;
} TrayIconInternal;

void cleanup (JNIEnv *env, jobject this, TrayIconInternal *internal);
DWORD ThreadBody (TrayIconInternal *internal);
BOOL CALLBACK WndProc (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp);

/////////////////////////////
// DLL initialization 
/////////////////////////////

static HINSTANCE hDLLInstance;

int APIENTRY DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{
    if (dwReason == DLL_PROCESS_ATTACH) {
        hDLLInstance = hInstance;
        DEBUG_PRINT1 ("hinstance = %d\n", hDLLInstance);
    }
    return 1;
}


/////////////////////////////
// Internal state
/////////////////////////////

TrayIconInternal* GetInternal (JNIEnv *env, jobject this);
void SetInternal (JNIEnv *env, jobject this, TrayIconInternal* internal);
TrayIconInternal* NewInternal (JNIEnv *env, jobject this);
void FreeInternal (JNIEnv *env, jobject this, TrayIconInternal *internal);

/*
 * Get internal pointer from Java object.
 */
TrayIconInternal* GetInternal (JNIEnv *env, jobject this)
{
    jclass cls;
    jfieldID field;

    // Get our class object
    cls = (*env)->GetObjectClass (env, this);
    //fprintf (stderr, "TrayIcon class=%p\n", cls);

    // Find the field ID for internal
    field = (*env)->GetFieldID (env, cls, "internal", "I");
    if (!field) {
        ERROR_PRINT0 ("can't find internal\n");
        return NULL;
    }

    return (TrayIconInternal*) (*env)->GetIntField (env, this, field);
}

/*
 * Set internal pointer in Java object.
 */
void SetInternal (JNIEnv *env, jobject this, TrayIconInternal* internal)
{
    jclass cls;
    jfieldID field;

    // Get our class object
    cls = (*env)->GetObjectClass (env, this);

    // Find the method ID for get()
    field = (*env)->GetFieldID (env, cls, "internal", "I");
    if (!field) {
        ERROR_PRINT0 ("can't find internal\n");
        return;
    }

    (*env)->SetIntField (env, this, field, (jint)internal);
}

/*
 * Make the internal data structure.
 */
TrayIconInternal* NewInternal (JNIEnv *env, jobject this)
{
    TrayIconInternal* internal;
    jclass cls;
    jfieldID field;

    internal = malloc (sizeof(*internal));
    if (!internal) {
        ERROR_PRINT0 ("Can't allocate internal structure\n");
        return NULL;
    }
    memset (internal, 0, sizeof(*internal));

    (*env)->GetJavaVM (env, &internal->vm);

    internal->this = (*env)->NewGlobalRef (env, this);
    DEBUG_PRINT2 ("this=%p, this=%p\n", this, internal->this);

    if (!internal->this) {
        ERROR_PRINT0 ("Can't create this\n");
        FreeInternal (env, this, internal);
        return NULL;
    }

    // Find the field and method IDs we'll need
    cls = (*env)->GetObjectClass (env, this);

    internal->internalMouseClickedID = 
        (*env)->GetMethodID (env, cls, "internalMouseClicked", "(ZZII)V");
    internal->visibleID = (*env)->GetFieldID (env, cls, "visible", "Z");
    internal->tooltipID = (*env)->GetFieldID (env, cls, "tooltip", "Ljava/lang/String;");

    // initialize the notify data structure
    memset(&internal->nid, 0, sizeof(internal->nid));
    internal->nid.cbSize = sizeof(internal->nid);
    internal->nid.uFlags = NIF_MESSAGE | NIF_TIP | NIF_ICON;
    internal->nid.uCallbackMessage = WM_USER;
    internal->nid.uID = 0;
    internal->nid.hWnd = 0;
    internal->nid.hIcon = LoadIcon (hDLLInstance, TEXT("TRAY_ICON"));

    // set the internal field
    SetInternal (env, this, internal);

    return internal;
}

void FreeInternal (JNIEnv *env, jobject this, TrayIconInternal *internal)
{
    // clear the internal field
    SetInternal (env, this, NULL);

    // Free the global TrayIcon ref
    (*env)->DeleteGlobalRef (env, internal->this);
    internal->this = NULL;

    // Detach this thread from the VM, making env invalid
    (*internal->vm)->DetachCurrentThread (internal->vm);
    internal->env = NULL;

    // free the internal structure
    free (internal);
}


/////////////////////////////
// Native methods
/////////////////////////////

/*
 * Class:     TrayIcon
 * Method:    create
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_rcm_win32_TrayIcon_create
  (JNIEnv *env, jobject this)
{
    TrayIconInternal *internal;
    DWORD idThread;

    DEBUG_PRINT0 ("TrayIcon_create\n");

    // save the internal pointer
    internal = NewInternal (env, this);
    if (!internal)
        return;

    // start the message-handling thread
    if (!CreateThread (NULL, 0, (LPTHREAD_START_ROUTINE)ThreadBody, 
                       (LPVOID)internal, 0, &idThread)) {
        ERROR_PRINT0 ("Can't create thread\n");
        FreeInternal (env, this, internal);
    }

    // wait until the hidden window is created
    while (!internal->hwnd) {
        Sleep (10);
    }
}


/*
 * Class:     TrayIcon
 * Method:    destroy
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_rcm_win32_TrayIcon_destroy
  (JNIEnv *env, jobject this)
{
    TrayIconInternal *internal;

    DEBUG_PRINT0 ("TrayIcon_destroy\n");

    internal = GetInternal (env, this);
    if (!internal)
        return;

    DestroyWindow (internal->hwnd);
}

/*
 * Class:     TrayIcon
 * Method:    changeVisibility
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_rcm_win32_TrayIcon_changeVisibility
  (JNIEnv *env, jobject this)
{
    TrayIconInternal *internal;
    jboolean visible;

    DEBUG_PRINT0 ("TrayIcon_changeVisibility\n");

    internal = GetInternal (env, this);

    visible = (*env)->GetBooleanField (env, this, internal->visibleID);

    DEBUG_PRINT1 ("TrayIcon: nid.hwnd = %p\n", internal->nid.hWnd);
    if (!Shell_NotifyIcon (visible ? NIM_ADD : NIM_DELETE, &internal->nid))
        ERROR_PRINT0 ("TrayIcon_changeVisibility: Shell_NotifyIcon failed\n");

    DEBUG_PRINT0 ("TrayIcon_changeVisibility: called Shell_NotifyIcon\n");
}

/*
 * Class:     TrayIcon
 * Method:    changeTooltip
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_rcm_win32_TrayIcon_changeTooltip
  (JNIEnv *env, jobject this)
{
    TrayIconInternal *internal;
    jstring tooltip;
    const char *strTooltip;
    jboolean visible;

    DEBUG_PRINT0 ("TrayIcon_changeTooltip\n");

    internal = GetInternal (env, this);

    tooltip = (*env)->GetObjectField (env, this, internal->tooltipID);

    if (!tooltip) {
        internal->nid.szTip[0] = 0;
    } else {
        strTooltip = (*env)->GetStringUTFChars (env, tooltip, 0);
        strcpy (internal->nid.szTip, strTooltip);
        (*env)->ReleaseStringUTFChars (env, tooltip, strTooltip);
    }

    visible = (*env)->GetBooleanField (env, this, internal->visibleID);

    if (visible) {
        if (!Shell_NotifyIcon (NIM_MODIFY, &internal->nid))
            ERROR_PRINT0 ("TrayIcon_changeTooltip: Shell_NotifyIcon failed\n");

        DEBUG_PRINT0 ("TrayIcon_changeTooltip: called Shell_NotifyIcon\n");
    }    
}


/////////////////////////////
// Hidden window
/////////////////////////////

/*
 * Message loop thread.
 */
DWORD ThreadBody (TrayIconInternal *internal)
{
    WNDCLASS wc;
    HWND hwnd;
    MSG msg;
    int result;

    DEBUG_PRINT0 ("TrayIcon_ThreadBody\n");

    // Attach this thread to the VM
    result = (*internal->vm)->AttachCurrentThread (internal->vm, (void*)&internal->env, NULL);
    DEBUG_PRINT2 ("AttachCurrentThread returned result=%d, env=%p\n", result, internal->env);

    wc.style = 0;
    wc.lpfnWndProc = (WNDPROC)WndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hIcon = NULL;
    wc.hInstance = hDLLInstance;
    wc.hCursor = NULL;
    wc.hbrBackground = (HBRUSH) GetStockObject( WHITE_BRUSH );
    wc.lpszMenuName = NULL;
    wc.lpszClassName = TEXT("rcm.win32.TrayIcon");

    if (!RegisterClass(&wc)) {
        ERROR_PRINT0 ("TrayIcon: can't register window class\n");
        FreeInternal (internal->env, internal->this, internal);
        return 0;
    }

    hwnd = CreateWindow (TEXT("rcm.win32.TrayIcon"),
                                   TEXT("rcm.win32.TrayIcon"),
                                   0,
                                   CW_USEDEFAULT,
                                   CW_USEDEFAULT,
                                   CW_USEDEFAULT,
                                   CW_USEDEFAULT,
                                   NULL, NULL, 
                                   hDLLInstance, (LPVOID)internal );
    if (!hwnd) {
        ERROR_PRINT0 ("TrayIcon: can't create window\n");
        FreeInternal (internal->env, internal->this, internal);
        return 0;
    }
    DEBUG_PRINT1 ("TrayIcon: hwnd = %p\n", hwnd);

    // store hwnd
    internal->nid.hWnd = hwnd;
    internal->hwnd = hwnd;

    // Message pump
    while (GetMessage (&msg, NULL, 0, 0)) {
        TranslateMessage (&msg);
        DispatchMessage (&msg);
    }

    FreeInternal (internal->env, internal->this, internal);
    return msg.wParam;
}

/*
 * Window procedure.
 */
BOOL CALLBACK WndProc (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
    TrayIconInternal *internal = 
        (TrayIconInternal*)GetWindowLong (hwnd, GWL_USERDATA);

    switch (msg) {
        case WM_CREATE:
            DEBUG_PRINT0 ("WM_CREATE\n");
            {
                CREATESTRUCT *cs = (CREATESTRUCT*)lp;
                internal = (TrayIconInternal *) cs->lpCreateParams;

                // Store the internal pointer
                SetWindowLong (hwnd, GWL_USERDATA, (LONG)internal);
            }
            break;

        case WM_DESTROY:
            DEBUG_PRINT0 ("WM_DESTROY\n");
            PostQuitMessage (0);
            break;
            
        case WM_USER:
            switch (lp) {
            case WM_LBUTTONUP:
            case WM_RBUTTONUP:
            case WM_LBUTTONDBLCLK:
            case WM_RBUTTONDBLCLK:
                DEBUG_PRINT0 ("WM_USER\n");
                {
                    POINT pos;
                    
                    GetCursorPos (&pos);
                    (*internal->env)->CallVoidMethod
                        (internal->env, internal->this, 
                         internal->internalMouseClickedID, 
                         (lp == WM_RBUTTONUP),
                         (lp == WM_LBUTTONDBLCLK || lp == WM_RBUTTONDBLCLK),
                         pos.x, pos.y);
                }
                break;
            }
            break;

        default:
            return DefWindowProc(hwnd, msg, wp, lp);
    }

    return TRUE;
}


