/*
 * Remote Clip synchronized clipboard
 * Copyright (C) 2000 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <process.h>
#include <io.h>

/*
 * Compile this file into an EXE with the desired name,
 * setting the following compile-time variables on
 * the compiler command line:
 *
 *     JAVA_ARGS   comma-separated list of strings: name of class to run
 *                 and other java args that should precede user's arguments
 *                 (e.g. -DJAVA_ARGS=\"-mx64\",\"RemoteClip\" )
 *                 The string *DIR* is replaced by 
 *                 the directory containing this executable.
 *
 *     CLASSPATH   comma-separated list of strings: items to insert at front
 *                 of CLASSPATH environment variable, for finding Java class
 *                 files.  The string *DIR* is replaced by the EXE's installation 
 *                 directory.  The string *.jar is replaced with all .jar files
 *                 found in the directory. You can use forward slashes for directory
 *                 separators.  For example, with this compile-time define:
 *                    -DCLASSPATH=\"../src\",\"../lib/*.jar\"
 *                 the following entries might be added to the CLASSPATH:
 *                        <EXE directory>/../src
 *                        <EXE directory>/../lib/rclip.jar
 *                        <EXE directory>/../lib/jsse.jar
 *                        etc.
 *
 *     PATH        comma-separated list of strings: items to insert at front
 *                 of PATH environment variable, in the same way as CLASSPATH.
 *                 As in CLASSPATH, *DIR* is replaced by the
 *                 EXE's installation directory.
 *
 *     CONSOLE     if exe should be a console app (otherwise it's a Windows
 *                 app with no access to stdin/stdout)
 */

#define DIRCODE   "*DIR*"

static char *classpathEntries[] = {
#ifdef CLASSPATH
  CLASSPATH,
#endif
  NULL
};

static char *pathEntries[] = {
#ifdef PATH
  PATH,
#endif
  NULL
};

static char *javaExe =
#ifdef CONSOLE
"java"
#else
"javaw"
#endif
;

static char *fixedArgs[] = {
#ifdef JAVA_ARGS
  JAVA_ARGS,
#endif
  NULL
};

#ifdef DEBUG
static int debug = 1;
#else
static int debug = 0;
#endif

// Check for -debug flag
void CheckDebugFlag (int argc, char **argv) {
    int i;
    for (i = 0; i < argc; ++i)
        if (!strcmp (argv[i], "-debug"))
            debug = 1;
}

void Dump (char *name, char **entries) {
  if (debug) {
    fprintf (stderr, "%s\n", name);
    while (*entries) {
        fprintf (stderr, "%s\n", *entries);
	++entries;
    }
  }
}
    

// Get directory containing this executable.
char *GetAppDir () 
{
  static char appDir[2048];

  if (!*appDir) {
    // Get the full pathname of the EXE file
    int len = GetModuleFileName (NULL, appDir, sizeof(appDir));
    
    // Strip off everything after the last slash,
    // leaving only the directory 
    char *s = strrchr (appDir, '\\');
    if (s)
      s[1] = '\0';
    else
      appDir[0] = '\0';
  }
  
  if (debug) {
      fprintf (stderr, "appdir = %s\n", appDir);
      fflush (stderr);
  }

  return appDir;
}

// Replace all occurrences of pat in string with subst
char *ReplaceAll (char *string, char *pat, char *subst) {
    char *p;
    while ((p = strstr(string, pat)) != NULL) {
        char *copy = malloc(strlen(string) + (strlen(subst) - strlen(pat)) + 1);
        strncpy (copy, string, p - string);
	copy[p-string] = 0;
        strcat (copy, subst);
        strcat (copy, p + strlen (pat));
        string = copy;
    }
    return string;
}

void ReplaceAllInArray (char **entries, char *pat, char *subst) {
    while (*entries) {
        *entries = ReplaceAll (*entries, pat, subst);
        ++entries;
    }
}

// convert forward slashes to backslashes
void FixSlashes (char **entries) {
  while (*entries) {
    char *p;
    while ((p = strchr (*entries, '/')) != NULL)
      *p = '\\';
    ++entries;
  }
}

int CalculatePathSpace (char **entries) {
  int size = 0;
  while (*entries) {
    size += strlen (*entries);
    ++entries;
  }
  return size;
}

void AppendPathEntries (char *result, char **entries) {
  while (*entries) {
    strcat (result, *entries);
    strcat (result, ";");
    ++entries;
  }
}

void ExtendPath (char *name, char **entries)
{
  char *currpath = getenv (name);
  char *result;

  if (currpath == NULL)
    currpath = "";

  FixSlashes (entries);

  result = malloc (strlen (name) 
                      + 1 /* equal sign */
		      + CalculatePathSpace (entries) + 1
		      + strlen (currpath));
  strcpy (result, name);
  strcat (result, "=");
  AppendPathEntries (result, entries);
  strcat (result, currpath);

  putenv (result);

  if (debug) {
      fprintf (stderr, "%s\n", result);
      fflush (stderr);
  }
}

// convert forward slashes to backslashes
void ForwardToBackSlashes (char *s) {
    while ((s = strchr (s, '/')) != NULL)
        *s = '\\';
}

char **FindFiles (char *filespec) {
  char **files;
  int maxFiles;
  int nFiles;
  char *dir;
  char *s;
  long hf;
  struct _finddata_t f;
  int i;
  
  ForwardToBackSlashes (filespec);
  dir = strdup (filespec);
  s = strrchr (dir, '\\');
  if (s)
    s[1] = '\0';
  else
    dir[0] = '\0';

  maxFiles = 10;
  files = malloc (sizeof(*files) * maxFiles);
  nFiles = 0;

  hf = _findfirst (filespec, &f);
  if (hf != -1) {
    do {
      int i = nFiles;

      if (++nFiles >= maxFiles) {
	maxFiles *= 2;
	files = realloc (files, sizeof (*files) * maxFiles);
      }

      files[i] = malloc (strlen (dir) + strlen (f.name));
      strcpy (files[i], dir);
      strcat (files[i], f.name);
#if DEBUG
      fprintf (stderr, "found: %s\n", files[i]);
#endif
    } while (_findnext (hf, &f) == 0);
    _findclose (hf);
  }

  files[nFiles] = NULL;
  return files;
}

char **FindAllFiles (char **entries) {
  char **files;
  int maxFiles;
  int nFiles;

  maxFiles = 10;
  files = malloc (sizeof(*files) * maxFiles);
  nFiles = 0;

  while (*entries) {
    char **f = FindFiles (*entries);
    while (*f) {
	int i = nFiles;
        if (++nFiles >= maxFiles) {
	  maxFiles *= 2;
  	  files = realloc (files, sizeof (*files) * maxFiles);
        }
	files[i] = *f;
	++f;
    }
    ++entries;
  }
  files[nFiles] = NULL;
  return files;
}

int CountArgs (char **args) {
  int n = 0;
  while (*args) {
    ++n;
    ++args;
  }
  return n;
}

char **MakeArgs (char *exename, char **fixedArgs, char **variableArgs) 
{
  int n = 2 + CountArgs (fixedArgs) + CountArgs (variableArgs);
  char **args = malloc (sizeof(char*) * n);
  int i = 0;
  int j;
  args[i++] = exename;
  for (j = 0; fixedArgs[j]; ++j)
    args[i++] = fixedArgs[j];
  for (j = 0; variableArgs[j]; ++j)
    args[i++] = variableArgs[j];
  args[i++] = NULL;
  return args;
}

int RunJava (char **javaArgs)
{
  int result;
  result = _spawnvp (_P_WAIT, javaArgs[0], javaArgs);
  if (result == -1 && errno != 0) {
    char *msg = strerror (errno);
    char *buf = malloc (20 + strlen (javaArgs[0]) + strlen (msg));
    sprintf (buf, "can't run %s: %s", javaArgs[0], msg);
#ifdef CONSOLE
    fprintf (stderr, "%s\n", buf);
#else
    MessageBox (NULL, buf, "Error", MB_OK | MB_ICONEXCLAMATION);
#endif
  }
  return result;
}

int main (int argc, char **argv)
{
  char **javaArgs;
  STARTUPINFO sinfo;
  PROCESS_INFORMATION pinfo;
  char *appdir = GetAppDir ();

  CheckDebugFlag (argc, argv);

  // Replace *DIR* with appdir
  ReplaceAllInArray (classpathEntries, DIRCODE, appdir);
  ReplaceAllInArray (pathEntries, DIRCODE, appdir);
  ReplaceAllInArray (fixedArgs, DIRCODE, appdir);

  // Extend environment variables 
  ExtendPath ("CLASSPATH", FindAllFiles (classpathEntries));
  ExtendPath ("PATH", pathEntries);

  // Set up Java command line.
  javaArgs = MakeArgs (javaExe, fixedArgs, argv+1);

  // Invoke Java.
  return RunJava (javaArgs);
}

#ifndef CONSOLE
//__declspec(dllimport) char **__initenv;

int WINAPI WinMain (HINSTANCE inst, HINSTANCE previnst, LPSTR cmdline, int cmdshow)
{
  //__initenv = _environ;
    return main (__argc, __argv);
}
#endif

